import re
from random import randint
from .color import tailwind_colors, radix_colors, semantic_colors

def is_digit(value):
    return value.isdigit()

def is_digit_range(value, minv, maxv):
    return value.isdigit() and minv <= int(value) <= maxv

def is_float(value):
    vs = value.split('.')
    return len(vs) == 2 and all((v=='' or v.isdigit()) for v in vs)

def is_fraction(value):
    vs = value.split('/')
    return len(vs) == 2 and all(v.isdigit() for v in vs)

def is_percent(value):
    return (len(value) > 1 and
            value[-1] == '%' and
            value[:-1].isdigit() and
            0 <= int(value[:-1]) <= 100)

def is_calc(value):
    return (len(value) > 6 and
            value[3] == ',' and
            value[:3] in ('add', 'sub','mul', 'div'))

def is_hex(value):
    return all('0'<=x<='9' or 'a'<=x<='f' or 'A'<=x<='F'
               for x in value)

def random_color(r=(0,255), g=(0,255), b=(0,255), a=255):
    def check(x):
        if (isinstance(r, (list, tuple)) and
            len(r) == 2 and
            0 <= r[0] <= r[1]):
            min = r[0]
            max = r[1] if r[1] <= 255 else 255
            return (min, max)
        elif isinstance(r, int):
            min = r if r >= 0 else 0
            min = min if min <= 255 else 255
            max = 255
            return (min, max)
        raise RuntimeError(f"Invalid argument '{x}'")
    r = randint(*check(r))
    g = randint(*check(g))
    b = randint(*check(b))
    a = randint(*check(a))
    return f'#{r:02x}{g:02x}{b:02x}{a:02x}'
    

def convert_size(value, negative=False):
    values = {
        '0':    '0px',
        'px':   '1px',
        'auto': 'auto',
        'full': '100%',
        'min':  'min-content',
        'max':  'max-content',
        'fit':  'fit-content',
        '3xs':  '16rem',       # 256px
        '2xs':  '18rem',       # 288px
        'xs':   '20rem',       # 320px
        'sm':   '24rem',       # 384px
        'md':   '28rem',       # 448px
        'lg':   '32rem',       # 512px
        'xl':   '36rem',       # 576px
        '2xl':  '42rem',       # 672px
        '3xl':  '48rem',       # 768px
        '4xl':  '56rem',       # 896px
        '5xl':  '64rem',       # 1024px
        '6xl':  '72rem',       # 1152px
        '7xl':  '80rem',       # 1280px
    }
    v = values.get(value)
    if v:
        size = v
    elif is_digit(value):
        size = f'{int(value)/4}rem'
    elif is_float(value):
        size = f'{float(value)/4}rem'
    elif is_fraction(value):
        a, b = value.split('/')
        a, b = int(a), int(b)
        size = f'{a*100/b}%'
    elif is_calc(value):
        items = value.split(',')
        if len(items) != 3:
            size = value
        else:
            ops = {'add':'+','sub':'-','mul':'*', 'div':'/'}
            size = f'calc({items[1]} {ops[items[0]]} {items[2]})'
    else:
        size = value
    if negative and size and size[0] != '-':
        size = '-' + size
    return size

def convert_color(args):
    if len(args) == 0:
        return None

    # 16进制格式的颜色，支持带有透明度
    if (len(args) == 2 and
        args[0] == 'hex' and
        is_hex(args[1]) and
        len(args[1]) in (3, 4, 6, 8)):
        return '#' + args[1]
    if (len(args) == 1 and
        args[0][0] == '#' and
        is_hex(args[0][1:]) and
        len(args[0]) in (4, 5, 7, 9)):
        return args[0]
    if (len(args) == 1 and
        is_hex(args[0]) and
        len(args[0]) in (3, 4, 6, 8)):
        return '#' + args[0]

    # tailwind风格的颜色，支持后跟'/75'这样添加透明度
    values = {
        'none':        'none',
        'inherit':     'inherit',
        'current':     'correntColor',
        'transparent': 'transparent',
        'black':       'rgb(0 0 0)',
        'white':       'rgb(255 255 255)',
    }
    value = '-'.join(args)
    color_opacity = value.split('/')
    if len(color_opacity) == 1:
        # 属性名中不能出现'/'符号，所以如果用无值属性时，应该用pink-900@25
        # 这样的方法，不能用pink-900/25这种方式，后者只能用于class值或属性值中
        # 2023-10-19: 当前fry语法中，属性名可以包含'/'，所以可以使用pink-900/25
        #             以下逻辑暂时保留
        color_opacity = value.split('@')
        if len(color_opacity) == 1:
            color = color_opacity[0]
            opacity = None
        else:
            color, opacity = color_opacity
    elif len(color_opacity) == 2:
        color, opacity = color_opacity
    else:
        return None
    c = (values.get(color) or
         tailwind_colors.get(color) or
         radix_colors.get(color) or
         semantic_colors.get(color))
    if not c:
        from fryhcs.css.plugin import plugin_color
        color = plugin_color(color)
        if not color:
            return None
        type = 'plugin'
    else:
        color = c
    if not opacity:
        return color
    if opacity.isdigit() and 0<=int(opacity)<=100:
        return re.sub(r'\)$', f' / {int(opacity)/100})', color)
    return None
        
def opacity_to(color, opacity):
    r = 255
    g = 255
    b = 255
    m = re.fullmatch(r'rgb\(([0-9]+) ([0-9]+) ([0-9]+)[^)]*\)', color)
    if color[0] == '#' and is_hex(color[1:]):
        if len(color) in (4, 5):
            r = int(int(color[1], 16)/15*255)
            g = int(int(color[2], 16)/15*255)
            b = int(int(color[3], 16)/15*255)
        elif len(color) in (7, 9):
            r = int(color[1:3], 16)
            g = int(color[3:5], 16)
            b = int(color[5:7], 16)
    elif m:
        r,g,b = m.groups()
    return f'rgb({r} {g} {b} / {opacity/100})'
        

def merge_value(args, negative=False):
    value = '-'.join(args)
    if negative and value and value[0] != '-':
        value = '-' + value
    return value.replace('_', ' ')


IGNORES = set(['css', 'args', 'argc', 'neg', 'from_plugin', 'utility_method', 'is_utility', 'add_style'])
class Utility():
    def __init__(self, css):
        args = css.utility_args
        self.css = css
        self.args = args
        self.argc = len(args)
        self.neg = len(args) > 0 and len(args[0]) > 0 and args[0][0] == '-'

    def from_plugin(self, pu):
        css = self.css
        css.plugin_order = pu.plugin_order
        css.level_order = pu.level_order
        css.selector_template = css.selector_template.format(selector=pu.selector_template)
        css.wrappers += pu.wrappers
        css.addons += pu.addons
        css.styles += pu.styles
        return True

    def utility_method(self):
        if self.argc == 0 or self.args[0] == '':
            return None

        from fryhcs.css.plugin import plugin_utility
        pu = plugin_utility(self.args)
        if pu:
            return lambda: self.from_plugin(pu)

        names = {
            'break': 'break_method',
            'not':   'not_method',
            'from':  'from_method',
        }
        name = self.args[0]

        if name in IGNORES:
            return None

        # 去掉负值的情况
        name = name[1:] if name[0] == '-' else name 

        # 处理python关键字
        name = names.get(name, name)

        return getattr(self, name, None)

    def is_utility(self):
        return True if self.utility_method() else False

    def __call__(self):
        method = self.utility_method()
        if not callable(method):
            return False
        return method()

    def add_style(self, key, value):
        self.css.add_style(key, value)

    def dummy(self):
        return True

    def aspect(self):
        if self.argc != 2:
            return False
        arg = self.args[1]
        values = {
            'auto':   'auto',
            'square': '1 / 1',
            'video':  '16 / 9'
        }
        value = values.get(arg, arg.replace('/', ' / '))
        self.add_style('aspect-ratio', str(value))
        return True

    def container(self):
        if self.argc != 1:
            return False
        self.add_style('width', '100%')
        for size in (640, 768, 1024, 1280, 1536):
            addon = self.css.new_addon()
            addon.wrappers.append(f'@media (min-width: {size}px)')
            addon.styles.append(('max-width', f'{size}px'))
        return True

    def columns(self):
        if self.argc != 2:
            return False
        arg = self.args[1]
        values = {
            '3xs': '16rem', # 256px
            '2xs': '18rem', # 288px
            'xs':  '20rem', # 320px
            'sm':  '24rem', # 384px
            'md':  '28rem', # 448px
            'lg':  '32rem', # 512px
            'xl':  '36rem', # 576px
            '2xl': '42rem', # 672px
            '3xl': '48rem', # 768px
            '4xl': '56rem', # 896px
            '5xl': '64rem', # 1024px
            '6xl': '72rem', # 1152px
            '7xl': '80rem', # 1280px
        }
        value = values.get(arg, arg)
        self.add_style('columns', str(value))
        return True

    def joined_args(self, values):
        arg = '-'.join(self.args)
        value = values.get(arg)
        if isinstance(value, list):
            for v in value:
                self.add_style(*v)
            return True
        elif isinstance(value, tuple):
            self.add_style(*value)
            return True
        return False

    def break_method(self):
        # how a column or page should break
        values = {
            'break-after-auto':          ('break-after', 'auto'),
            'break-after-avoid':         ('break-after', 'avoid'),
            'break-after-all':           ('break-after', 'all'),
            'break-after-avoid-page':    ('break-after', 'avoid-page'),
            'break-after-page':          ('break-after', 'page'),
            'break-after-left':          ('break-after', 'left'),
            'break-after-right':         ('break-after', 'right'),
            'break-after-column':        ('break-after', 'column'),
            'break-before-auto':         ('break-before', 'auto'),
            'break-before-avoid':        ('break-before', 'avoid'),
            'break-before-all':          ('break-before', 'all'),
            'break-before-avoid-page':   ('break-before', 'avoid-page'),
            'break-before-page':         ('break-before', 'page'),
            'break-before-left':         ('break-before', 'left'),
            'break-before-right':        ('break-before', 'right'),
            'break-before-column':       ('break-before', 'column'),
            'break-inside-auto':         ('break-inside', 'auto'),
            'break-inside-avoid':        ('break-inside', 'avoid'),
            'break-inside-avoid-page':   ('break-inside', 'avoid-page'),
            'break-inside-avoid-column': ('break-inside', 'avoid-column'),
        }
        if self.joined_args(values):
            return True

        values = {
            'break-normal': [('overflow-wrap', 'normal'), ('word-break', 'normal')],
            'break-words':  ('overflow-wrap', 'break-word'),
            'break-all':    ('word-break', 'break-all'),
            'break-keep':   ('word-break', 'keep-all'),
        }
        return self.joined_args(values)

    def box(self):
        values = {
            'box-decoration-clone': ('box-decoration-break', 'clone'),
            'box-decoration-slice': ('box-decoration-break', 'slice'),
            'box-border':           ('box-sizing', 'border-box'),
            'box-content':          ('box-sizing', 'content-box'),
        }
        return self.joined_args(values)

    def block(self):
        return self.joined_args({'block': ('display', 'block')})

    def inline(self):
        values = {
            'inline':       ('display', 'inline'),
            'inline-block': ('display', 'inline-block'),
            'inline-flex':  ('display', 'inline-flex'),
            'inline-table': ('display', 'inline-table'),
        }
        return self.joined_args(values)

    def flex(self):
        values = {
            'flex':              ('display', 'flex'),
            'flex-row':          ('flex-direction', 'row'),
            'flex-row-reverse':  ('flex-direction', 'row-reverse'),
            'flex-col':          ('flex-direction', 'column'),
            'flex-col-reverse':  ('flex-direction', 'column-reverse'),
            'flex-wrap':         ('flex-wrap', 'wrap'),
            'flex-wrap-reverse': ('flex-wrap', 'wrap-reverse'),
            'flex-nowrap':       ('flex-wrap', 'nowrap'),
            'flex-1':            ('flex', '1 1 0%'),
            'flex-auto':         ('flex', '1 1 auto'),
            'flex-initial':      ('flex', '0 1 auto'),
            'flex-none':         ('flex', 'none'),
        }
        if self.joined_args(values):
            return True
        if self.argc != 4:
            return False
        if self.args[1].isdigit() and self.args[2].isdigit():
            self.add_style('flex', f'{self.args[1]} {self.args[2]} {self.args[3]}')
            return True
        return False

    def table(self):
        # set table-related display
        values = {
            'table':              ('display', 'table'),
            'table-caption':      ('display', 'table-caption'),
            'table-cell':         ('display', 'table-cell'),
            'table-column':       ('display', 'table-column'),
            'table-column-group': ('display', 'table-column-group'),
            'table-footer-group': ('display', 'table-footer-group'),
            'table-header-group': ('display', 'table-header-group'),
            'table-row-group':    ('display', 'table-row-group'),
            'table-row':          ('display', 'table-row'),
        }
        if self.joined_args(values):
            return True

        # set table layout
        values = {
            'table-auto':  ('table-layout', 'auto'),
            'table-fixed': ('table-layout', 'fixed'),
        }
        return self.joined_args(values)

    def flow(self):
        return self.joined_args({'flow-root': ('display', 'flow-root')})

    def grid(self):
        values = {
            'grid':                ('display', 'grid'),
            'grid-cols-none':      ('grid-template-columns', 'none'),
            'grid-rows-none':      ('grid-template-rows', 'none'),
            'grid-flow-row':       ('grid-auto-flow', 'row'),
            'grid-flow-col':       ('grid-auto-flow', 'column'),
            'grid-flow-dense':     ('grid-auto-flow', 'dense'),
            'grid-flow-row-dense': ('grid-auto-flow', 'row dense'),
            'grid-flow-col-dense': ('grid-auto-flow', 'column dense'),
        }

        if self.joined_args(values):
            return True

        if self.argc >= 3 and self.args[1] == 'cols':
            if self.argc == 3 and self.args[2].isdigit():
                value = f'repeat({self.args[2]}, minmax(0, 1fr))'
            else:
                value = merge_value(self.args[2:])
            self.add_style('grid-template-columns',  value)
            return True
        elif self.argc >= 3 and self.args[1] == 'rows':
            if self.argc == 3 and self.args[2].isdigit():
                value = f'repeat({self.args[2]}, minmax(0, 1fr))'
            else:
                value = merge_value(self.args[2:])
            self.add_style('grid-template-rows', value)
            return True


    def contents(self):
        return self.joined_args({'contents': ('display', 'contents')})

    def list(self):
        if self.argc == 1:
            return False
        values = {
            'list-item':       ('display', 'list-item'),
            'list-image-none': ('list-style-image', 'none'),
            'list-inside':     ('list-style-position', 'inside'),
            'list-outside':    ('list-style-position', 'outside'),
            'list-none':       ('list-style-type', 'none'),
            'list-disc':       ('list-style-type', 'disc'),
            'list-decimal':    ('list-style-type', 'decimal'),
        }
        if self.joined_args(values):
            return True
        if self.args[1] == 'image':
            value = merge_value(self.args[2:])
            self.add_style('list-style-image', value)
        else:
            value = merge_value(self.args[1:])
            self.add_style('list-style-type', value)
        return True

    def hidden(self):
        return self.joined_args({'hidden': ('display', 'none')})

    def float(self):
        values = {
            'float-right': ('float', 'right'),
            'float-left':  ('float', 'left'),
            'float-none':  ('float', 'none'),
        }
        return self.joined_args(values)

    def clear(self):
        values = {
            'clear-left':  ('clear', 'left'),
            'clear-right': ('clear', 'right'),
            'clear-both':  ('clear', 'both'),
            'clear-none':  ('clear', 'none'),
        }
        return self.joined_args(values)

    def isolate(self):
        return self.joined_args({'isolate': ('isolation', 'isolate')})

    def isolation(self):
        return self.joined_args({'isolation-auto': ('isolation', 'auto')})

    def object(self):
        values = {
            'object-contain':      ('object-fit', 'contain'),
            'object-cover':        ('object-fit', 'cover'),
            'object-fill':         ('object-fit', 'fill'),
            'object-none':         ('object-fit', 'none'),
            'object-scale-down':   ('object-fit', 'scale-down'),
            'object-bottom':       ('object-position', 'bottom'),
            'object-center':       ('object-position', 'center'),
            'object-left':         ('object-position', 'left'),
            'object-left-bottom':  ('object-position', 'left bottom'),
            'object-left-top':     ('object-position', 'left top'),
            'object-right':        ('object-position', 'right'),
            'object-right-bottom': ('object-position', 'right bottom'),
            'object-right-top':    ('object-position', 'right top'),
            'object-top':          ('object-position', 'top'),
        }
        return self.joined_args(values)

    def overflow(self):
        values = {
            'overflow-auto':      ('overflow', 'auto'),
            'overflow-hidden':    ('overflow', 'hidden'),
            'overflow-clip':      ('overflow', 'clip'),
            'overflow-visible':   ('overflow', 'visible'),
            'overflow-scroll':    ('overflow', 'scroll'),
            'overflow-x-auto':    ('overflow-x', 'auto'),
            'overflow-x-hidden':  ('overflow-x', 'hidden'),
            'overflow-x-clip':    ('overflow-x', 'clip'),
            'overflow-x-visible': ('overflow-x', 'visible'),
            'overflow-x-scroll':  ('overflow-x', 'scroll'),
            'overflow-y-auto':    ('overflow-y', 'auto'),
            'overflow-y-hidden':  ('overflow-y', 'hidden'),
            'overflow-y-clip':    ('overflow-y', 'clip'),
            'overflow-y-visible': ('overflow-y', 'visible'),
            'overflow-y-scroll':  ('overflow-y', 'scroll'),
        }
        return self.joined_args(values)

    def overscroll(self):
        values = {
            'overscroll-auto':      ('overscroll-behavior', 'auto'),
            'overscroll-contain':   ('overscroll-behavior', 'contain'),
            'overscroll-none':      ('overscroll-behavior', 'none'),
            'overscroll-x-auto':    ('overscroll-behavior-x', 'auto'),
            'overscroll-x-contain': ('overscroll-behavior-x', 'contain'),
            'overscroll-x-none':    ('overscroll-behavior-x', 'none'),
            'overscroll-y-auto':    ('overscroll-behavior-y', 'auto'),
            'overscroll-y-contain': ('overscroll-behavior-y', 'contain'),
            'overscroll-y-none':    ('overscroll-behavior-y', 'none'),
        }
        return self.joined_args(values)

    def static(self):
        return self.joined_args({'static': ('position', 'static')})

    def fixed(self):
        return self.joined_args({'fixed': ('position', 'fixed')})

    def absolute(self):
        return self.joined_args({'absolute': ('position', 'absolute')})

    def relative(self):
        return self.joined_args({'relative': ('position', 'relative')})

    def sticky(self):
        return self.joined_args({'sticky': ('position', 'sticky')})

    def inset(self):
        if self.argc == 2:
            direction = None
            value = self.args[1]
        elif self.argc == 3:
            direction = self.args[1]
            value = self.args[2]
            if direction != 'x' or direction != 'y':
                return False
        else:
            return False
        value = convert_size(value, self.neg)
        if not direction:
            self.add_style('inset', value)
        elif direction == 'x':
            self.add_style('left', value)
            self.add_style('right', value)
        else:
            self.add_style('top', value)
            self.add_style('bottom', value)
        return True

    def start(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1], self.neg)
        self.add_style('inset-inline-start', value)
        return True

    def end(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1], self.neg)
        self.add_style('inset-inline-end', value)
        return True

    def top(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1], self.neg)
        self.add_style('top', value)
        return True
    
    def right(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1], self.neg)
        self.add_style('right', value)
        return True

    def bottom(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1], self.neg)
        self.add_style('bottom', value)
        return True

    def left(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1], self.neg)
        self.add_style('left', value)
        return True

    def visible(self):
        return self.joined_args({'visible': ('visibility', 'visible')})

    def invisible(self):
        return self.joined_args({'invisible': ('visibility', 'hidden')})

    def collapse(self):
        return self.joined_args({'collapse': ('visibility', 'collapse')})

    def z(self):
        if self.argc != 2:
            return False
        if self.args[0][0] == '-':
            value = '-' + self.args[1]
        else:
            value = self.args[1]
        self.add_style('z-index', value)
        return True

    def basis(self):
        if self.argc != 2:
            return False
        value = convert_size(self.args[1])
        self.add_style('flex-basis', value)
        return True

    def grow(self):
        if self.argc == 1:
            value = 1
        elif self.argc != 2:
            return False
        elif not self.args[1].isdigit():
            return False
        else:
            value = self.args[1]
        self.add_style('flex-grow', value)
        return True

    def shrink(self):
        if self.argc == 1:
            value = 1
        elif self.argc != 2:
            return False
        elif not self.args[1].isdigit():
            return False
        else:
            value = self.args[1]
        self.add_style('flex-shrink', value)
        return True

    def order(self):
        if self.argc != 2:
            return False
        elif self.args[1] == 'first':
            value = '-9999'
        elif self.args[1] == 'last':
            value = '9999'
        elif self.args[1] == 'none':
            value = '0'
        elif not self.args[1].isdigit():
            return False
        else:
            value = self.args[1]
            if self.args[0][0] == '-':
                value = '-' + value
        self.add_style('order', value)
        return True

    def col(self):
        values = {
            'col-auto':       ('grid-column', 'auto'),
            'col-span-full':  ('grid-column', '1 / -1'),
            'col-start-auto': ('grid-column-start', 'auto'),
            'col-end-auto':   ('grid-column-end', 'auto'),
        }

        if self.joined_args(values):
            return True
        if self.argc == 3 and self.args[1] == 'span' and self.args[2].isdigit():
            name, value = 'grid-column', f'span {self.args[2]} / span {self.args[2]}'
        elif self.argc == 3 and self.args[1] in ('start', 'end') and self.args[2].isdigit():
            name, value = f'grid-column-{self.args[1]}', self.args[2]
        else:
            name, value = 'grid-column', merge_value(self.args[1:])
        self.add_style(name, value)
        return True

    def row(self):
        values = {
            'row-auto':       ('grid-row', 'auto'),
            'row-span-full':  ('grid-row', '1 / -1'),
            'row-start-auto': ('grid-row-start', 'auto'),
            'row-end-auto':   ('grid-row-end', 'auto'),
        }

        if self.joined_args(values):
            return True
        if self.argc == 3 and self.args[1] == 'span' and self.args[2].isdigit():
            name, value = 'grid-row', f'span {self.args[2]} / span {self.args[2]}'
        elif self.argc == 3 and self.args[1] in ('start', 'end') and self.args[2].isdigit():
            name, value = f'grid-row-{self.args[1]}', self.args[2]
        else:
            name, value = 'grid-row', merge_value(self.args[1:])
        self.add_style(name, value)
        return True

    def auto(self):
        values = {
            'auto-cols-auto': ('grid-auto-columns', 'auto'),
            'auto-cols-min':  ('grid-auto-columns', 'min-content'),
            'auto-cols-max':  ('grid-auto-columns', 'max-content'),
            'auto-cols-fr':   ('grid-auto-columns', 'minmax(0, 1fr)'),
            'auto-rows-auto': ('grid-auto-rows', 'auto'),
            'auto-rows-min':  ('grid-auto-rows', 'min-content'),
            'auto-rows-max':  ('grid-auto-rows', 'max-content'),
            'auto-rows-fr':   ('grid-auto-rows', 'minmax(0, 1fr)'),
        }
        if self.joined_args(values):
            return True
        if self.argc >= 3 and self.args[1] in ('cols', 'rows'):
            arg = self.args[1]
            args = self.args[2:]
            name, value = f"grid-auto-{'columns' if arg == 'cols' else arg}", merge_value(args)
            self.add_style(name, value)
            return True
        return False

    def gap(self):
        if self.argc == 2:
            direction = None
            value = self.args[1]
        elif self.argc == 3:
            if self.args[1] not in ('x', 'y'):
                return False
            direction = self.args[1]
            value = self.args[2]
        else:
            return False
        # tailwind不支持负的gap
        value = convert_size(value)
        if not direction:
            self.add_style('gap', value)
        elif direction == 'x':
            self.add_style('column-gap', value)
        else:
            self.add_style('row-gap', value)
        return True

    def justify(self):
        values = {
            'justify-normal':        ('justify-content', 'normal'),
            'justify-start':         ('justify-content', 'flex-start'),
            'justify-end':           ('justify-content', 'flex-end'),
            'justify-center':        ('justify-content', 'center'),
            'justify-between':       ('justify-content', 'space-between'),
            'justify-around':        ('justify-content', 'space-around'),
            'justify-evenly':        ('justify-content', 'space-evenly'),
            'justify-stretch':       ('justify-content', 'stretch'),
            'justify-items-start':   ('justify-items', 'start'),
            'justify-items-end':     ('justify-items', 'end'),
            'justify-items-center':  ('justify-items', 'center'),
            'justify-items-stretch': ('justify-items', 'stretch'),
            'justify-self-auto':     ('justify-self', 'auto'),
            'justify-self-start':    ('justify-self', 'start'),
            'justify-self-end':      ('justify-self', 'end'),
            'justify-self-center':   ('justify-self', 'center'),
            'justify-self-stretch':  ('justify-self', 'stretch'),
        }
        return self.joined_args(values)

    def content(self):
        # align content
        values = {
            'content-normal':   ('align-content', 'normal'),
            'content-center':   ('align-content', 'center'),
            'content-start':    ('align-content', 'flex-start'),
            'content-end':      ('align-content', 'flex-end'),
            'content-between':  ('align-content', 'space-between'),
            'content-around':   ('align-content', 'space-around'),
            'content-evenly':   ('align-content', 'space-evenly'),
            'content-baseline': ('align-content', 'baseline'),
            'content-stretch':  ('align-content', 'stretch'),
        }
        if self.joined_args(values):
            return True

        # content of the before or after pseudo-elements
        values = {
            'content-none': ('content', 'none'),
        }
        if self.joined_args(values):
            return True

        # arbitrary content
        if self.argc > 1:
            value = merge_value(self.args[1:])
            self.add_style('content', value)
            return True

        return False

    def items(self):
        values = {
            'items-start':    ('align-items', 'flex-start'),
            'items-end':      ('align-items', 'flex-end'),
            'items-center':   ('align-items', 'center'),
            'items-baseline': ('align-items', 'baseline'),
            'items-stretch':  ('align-items', 'stretch'),
        }
        return self.joined_args(values)

    def self(me):
        values = {
            'self-auto':     ('align-self', 'auto'),
            'self-start':    ('align-self', 'flex-start'),
            'self-end':      ('align-self', 'flex-end'),
            'self-center':   ('align-self', 'center'),
            'self-baseline': ('align-self', 'baseline'),
            'self-stretch':  ('align-self', 'stretch'),
        }
        return me.joined_args(values)

    def place(self):
        values = {
            'place-content-center':   ('place-content', 'center'),
            'place-content-start':    ('place-content', 'start'),
            'place-content-end':      ('place-content', 'end'),
            'place-content-between':  ('place-content', 'space-between'),
            'place-content-around':   ('place-content', 'space-around'),
            'place-content-evenly':   ('place-content', 'space-evenly'),
            'place-content-baseline': ('place-content', 'baseline'),
            'place-content-stretch':  ('place-content', 'stretch'),
            'place-items-start':      ('place-items', 'start'),
            'place-items-end':        ('place-items', 'end'),
            'place-items-center':     ('place-items', 'center'),
            'place-items-baseline':   ('place-items', 'baseline'),
            'place-items-stretch':    ('place-items', 'stretch'),
            'place-self-auto':        ('place-self', 'auto'),
            'place-self-start':       ('place-self', 'start'),
            'place-self-end':         ('place-self', 'end'),
            'place-self-center':      ('place-self', 'center'),
            'place-self-stretch':     ('place-self', 'stretch'),
        }
        return self.joined_args(values)

    def p(self):
        if self.argc == 2:
            size = convert_size(self.args[1])
            self.add_style('padding', size)
            return True
        elif self.argc == 3 and self.args[1] in 'xysetrbl':
            self.args = [self.args[0]+self.args[1], self.args[2]]
            self.argc = 2
            return self()
        return False

    def px(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-left', size)
        self.add_style('padding-right', size)
        return True

    def py(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-top', size)
        self.add_style('padding-bottom', size)
        return True

    def ps(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-inline-start', size)
        return True

    def pe(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-inline-end', size)
        return True

    def pt(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-top', size)
        return True

    def pr(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-right', size)
        return True

    def pb(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-bottom', size)
        return True

    def pl(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1])
        self.add_style('padding-left', size)
        return True

    def m(self):
        if self.argc == 2:
            size = convert_size(self.args[1], self.neg)
            self.add_style('margin', size)
            return True
        elif self.argc == 3 and self.args[1] in 'xysetrbl':
            self.args = [self.args[0]+self.args[1], self.args[2]]
            self.argc = 2
            return self()
        return False

    def mx(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-left', size)
        self.add_style('margin-right', size)
        return True

    def my(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-top', size)
        self.add_style('margin-bottom', size)
        return True

    def ms(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-inline-start', size)
        return True

    def me(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-inline-end', size)
        return True

    def mt(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-top', size)
        return True

    def mr(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-right', size)
        return True

    def mb(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-bottom', size)
        return True

    def ml(self):
        if self.argc != 2:
            return False
        size = convert_size(self.args[1], self.neg)
        self.add_style('margin-left', size)
        return True

    def space(self):
        if self.argc != 3 or self.args[1] not in 'xy':
            return False
        self.css.selector_template += ' > :not([hidden]) ~ :not([hidden])'

        dir = self.args[1]
        if self.args[2] == 'reverse':
            self.add_style(f'--fry-space-{dir}-reverse', '1')
        else:
            size = convert_size(self.args[2], self.neg)
            if dir == 'x':
                self.add_style('--fry-space-x-reverse',  '0')
                self.add_style('margin-right', f'calc({size} * var(--fry-space-x-reverse))')
                self.add_style('margin-left', f'calc({size} * calc(1 - var(--fry-space-x-reverse)))')
            else:
                self.add_style('--fry-space-y-reverse', '0')
                self.add_style('margin-bottom', f'calc({size} * var(--fry-space-y-reverse))')
                self.add_style('margin-top', f'calc({size} * calc(1 - var(--fry-space-y-reverse)))')
        return True

    def w(self):
        if self.argc != 2:
            return False
        if self.joined_args({'w-screen': ('width', '100vw')}):
            return True
        size = convert_size(self.args[1])
        self.add_style('width', size)
        return True

    def h(self):
        if self.argc != 2:
            return False
        if self.joined_args({'h-screen': ('height', '100vh')}):
            return True
        size = convert_size(self.args[1])
        self.add_style('height', size)
        return True

    def min(self):
        if self.joined_args({'min-h-screen': ('min-height', '100vh')}):
            return True

        if self.argc != 3 or self.args[1] not in 'hw':
            return False

        type = 'height' if self.args[1] == 'h' else 'width'
        size = convert_size(self.args[2])
        self.add_style(f'min-{type}', size)
        return True
        
    def max(self):
        values = {
            'max-w-none':       ('max-width', 'none'),
            'max-w-prose':      ('max-width', '65ch'),
            'max-w-screen-sm':  ('max-width', '640px'),
            'max-w-screen-md':  ('max-width', '768px'),
            'max-w-screen-lg':  ('max-width', '1024px'),
            'max-w-screen-xl':  ('max-width', '1280px'),
            'max-w-screen-2xl': ('max-width', '1536px'),
            'max-h-none':       ('max-height', 'none'),
            'max-h-screen':     ('max-height', '100vh'),
        }
        if self.joined_args(values):
            return True

        if self.argc != 3 or self.args[1] not in 'hw':
            return False

        type = 'height' if self.args[1] == 'h' else 'width'
        size = convert_size(self.args[2])
        self.add_style(f'max-{type}', size)
        return True

    def font(self):
        # font family, font weight
        values = {
            'font-sans':       ('font-family', 'ui-sans-serif, system-ui, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue", Arial, "Noto Sans", sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol", "Noto Color Emoji"'),
            'font-serif':      ('font-family', 'ui-serif, Georgia, Cambria, "Times New Roman", Times, serif'),
            'font-mono':       ('font-family', 'ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas,"Liberation Mono", "Courier New", monospace'),
            'font-thin':       ('font-weight', '100'),
            'font-extralight': ('font-weight', '200'),
            'font-light':      ('font-weight', '300'),
            'font-normal':     ('font-weight', '400'),
            'font-medium':     ('font-weight', '500'),
            'font-semibold':   ('font-weight', '600'),
            'font-bold':       ('font-weight', '700'),
            'font-extrabold':  ('font-weight', '800'),
            'font-black':      ('font-weight', '900'),
        }
        if self.joined_args(values):
            return True
        value = merge_value(self.args[1:])
        if value.isdigit():
            self.add_style('font-weight', value)
        else:
            self.add_style('font-family', value)
        return True

    def text(self):
        # text overflow
        values = {
            'text-ellipsis': ('text-overflow', 'ellipsis'),
            'text-clip':     ('text-overflow', 'clip'),
        }

        if self.joined_args(values):
            return True

        # text align
        values = {
            'text-left':        ('text-align', 'left'),
            'text-center':      ('text-align', 'center'),
            'text-right':       ('text-align', 'right'),
            'text-justify':     ('text-align', 'justify'),
            'text-start':       ('text-align', 'start'),
            'text-end':         ('text-align', 'end'),
        }

        if self.joined_args(values):
            return True

        if self.argc < 2:
            return False

        # text color
        color = convert_color(self.args[1:])

        if color:
            self.add_style('color', color)
            return True

        # font size
        font_sizes = {
            'xs': ('0.75rem', '1rem'),
            'sm': ('0.875rem', '1.25rem'),
            'base': ('1rem', '1.5rem'),
            'lg': ('1.125rem', '1.75rem'),
            'xl': ('1.25rem', '1.75rem'),
            '2xl': ('1.5rem', '2rem'),
            '3xl': ('1.875rem', '2.25rem'),
            '4xl': ('2.25rem', '2.5rem'),
            '5xl': ('3rem', '1'),
            '6xl': ('3.75rem', '1'),
            '7xl': ('4.5rem', '1'),
            '8xl': ('6rem', '1'),
            '9xl': ('8rem', '1'),
        }

        if self.argc == 2 and self.args[1] in font_sizes:
            fs, lh = font_sizes[self.args[1]]
            self.add_style('font-size', fs)
            self.add_style('line-height', lh)
            return True

        # arbitrary font size
        # e.g. text-14px
        size = merge_value(self.args[1:])
        self.add_style('font-size', size)
        return True

    def antialiased(self):
        if self.argc != 1:
            return False
        self.add_style('-webkit-font-smoothing', 'antialiased')
        self.add_style('-moz-osx-font-smoothing', 'grayscale')
        return True

    def subpixel(self):
        if self.argc == 2 and self.args[1] == 'antialiased':
            self.add_style('-webkit-font-smoothing', 'auto')
            self.add_style('-moz-osx-font-smoothing', 'auto')
            return True
        return False

    def italic(self):
        if self.argc != 1:
            return False
        self.add_style('font-style', 'italic')
        return True

    def not_method(self):
        values = {
            'not-italic':   ('font-style', 'normal'),
            'not-sr-only':  [('position', 'static'),
                             ('width', 'auto'),
                             ('height', 'auto'),
                             ('padding', '0'),
                             ('margin', '0'),
                             ('overflow', 'visible'),
                             ('clip', 'auto'),
                             ('white-space', 'normal')],
        }
        return self.joined_args(values)

    def normal(self):
        # font-variant-numeric and text-transform
        values = {
            'normal-nums': ('font-variant-numeric', 'normal'),
            'normal-case': ('text-transform', 'none'),
        }
        return self.joined_args(values)

    def ordinal(self):
        values = {
            'ordinal': ('font-variant-numeric', 'ordinal'),
        }
        return self.joined_args(values)

    def slashed(self):
        values = {
            'slashed-zero': ('font-variant-numeric', 'slashed-zero'),
        }
        return self.joined_args(values)

    def lining(self):
        values = {
            'lining-nums': ('font-variant-numeric', 'lining-nums'),
        }
        return self.joined_args(values)

    def oldstyle(self):
        values = {
            'oldstyle-nums': ('font-variant-numeric', 'oldstyle-nums'),
        }
        return self.joined_args(values)

    def proportional(self):
        values = {
            'proportional-nums': ('font-variant-numeric', 'proportional-nums'),
        }
        return self.joined_args(values)

    def tabular(self):
        values = {
            'tabular-nums': ('font-variant-numeric', 'tabular-nums'),
        }
        return self.joined_args(values)

    def diagonal(self):
        values = {
            'diagonal-fractions': ('font-variant-numeric', 'diagonal-fractions'),
        }
        return self.joined_args(values)

    def stacked(self):
        values = {
            'stacked-fractions': ('font-variant-numeric', 'stacked-fractions'),
        }
        return self.joined_args(values)

    def tracking(self):
        if self.argc != 2:
            return False

        letter_spacings = {
            'tighter': '-0.05em',
            'tight':   '-0.025em',
            'normal':  '0em',
            'wide':    '0.025em',
            'wider':   '0.05em',
            'widest':  '0.1em',
        }

        ls = letter_spacings.get(self.args[1], self.args[1])
        if self.neg and ls and ls[0] != '-':
            ls = '-' + ls
        self.add_style('letter-spacing', ls)
        return True

    def line(self):
        # text decoration: line-through
        values = {
            'line-through': ('text-decoration-line', 'line-through'),
        }
        if self.joined_args(values):
            return True

        # line clamp
        if self.argc != 3 or self.args[1] != 'clamp':
            return False
        if self.args[2] == 'none':
            self.add_style('overflow', 'visible')
            self.add_style('display', 'block')
            self.add_style('-webkit-box-orient', 'horizontal')
            self.add_style('-webkit-line-clamp', 'none')
            return True
        if self.args[2].isdigit():
            self.add_style('overflow', 'hidden')
            self.add_style('display', '-webkit-box')
            self.add_style('-webkit-box-orient', 'vertical')
            self.add_style('-webkit-line-clamp', self.args[2])
            return True
        return False

    def leading(self):
        if self.argc != 2:
            return False
        line_heights = {
            'none': '1',
            'tight': '1.25',
            'snug': '1.375',
            'normal': '1.5',
            'relaxed': '1.625',
            'loose': '2',
        }
        height = line_heights.get(self.args[1], convert_size(self.args[1]))
        self.add_style('line-height', height)
        return True

    def underline(self):
        # text-decoration-line: underline
        if self.argc == 1:
            self.add_style('text-decoration-line', 'underline')
            return True

        # text-underline-offset
        values = {
            'underline-offset-auto': ('text-underline-offset', 'auto'),
            'underline-offset-0':    ('text-underline-offset', '0px'),
            'underline-offset-1':    ('text-underline-offset', '1px'),
            'underline-offset-2':    ('text-underline-offset', '2px'),
            'underline-offset-4':    ('text-underline-offset', '4px'),
            'underline-offset-8':    ('text-underline-offset', '8px'),
        }
        if self.joined_args(values):
            return True

        if self.argc > 2 and self.args[1] == 'offset':
            value = merge_value(self.args[2:])
            self.add_style('text-underline-offset', value)
            return True

    def overline(self):
        if self.argc != 1:
            return False
        self.add_style('text-decoration-line', 'overline')
        return True

    def no(self):
        values = {
            'no-underline': ('text-decoration-line', 'none'),
        }
        return self.joined_args(values)

    def decoration(self):
        if self.argc < 2:
            return False
        # text decoration style
        values = {
            'decoration-solid':  ('text-decoration-style', 'solid'),
            'decoration-double': ('text-decoration-style', 'double'),
            'decoration-dotted': ('text-decoration-style', 'dotted'),
            'decoration-dashed': ('text-decoration-style', 'dashed'),
            'decoration-wavy':   ('text-decoration-style', 'wavy'),
        }
        if self.joined_args(values):
            return True

        # text decoration color
        color = convert_color(self.args[1:])
        if color:
            self.add_style('text-decoration-color', color)
            return True

        # text decoration thickness
        values = {
            'decoration-auto':      ('text-decoration-thickness', 'auto'),
            'decoration-from-font': ('text-decoration-thickness', 'from-font'),
            'decoration-0':         ('text-decoration-thickness', '0px'),
            'decoration-1':         ('text-decoration-thickness', '1px'),
            'decoration-2':         ('text-decoration-thickness', '2px'),
            'decoration-4':         ('text-decoration-thickness', '4px'),
            'decoration-8':         ('text-decoration-thickness', '8px'),
        }
        if self.joined_args(values):
            return True

        value = merge_value(self.args[1:])
        self.add_style('text-decoration-thickness', value)
        return True

    def uppercase(self):
        if self.argc == 1:
            self.add_style('text-transform', 'uppercase')
            return True
        return False

    def lowercase(self):
        if self.argc == 1:
            self.add_style('text-transform', 'lowercase')
            return True
        return False

    def capitalize(self):
        if self.argc == 1:
            self.add_style('text-transform', 'capitalize')
            return True
        return False

    def truncate(self):
        if self.argc == 1:
            self.add_style('overflow', 'hidden')
            self.add_style('text-overflow', 'ellipsis')
            self.add_style('white-space', 'nowrap')
            return True
        return False

    def indent(self):
        if self.argc == 1:
            return False
        elif self.argc == 2:
            size = convert_size(self.args[1], self.neg)
        else:
            size = merge_value(self.args[1:], self.neg)
        self.add_style('text-indent', size)
        return True

    def align(self):
        values = {
            'align-baseline':    ('vertical-align', 'baseline'),
            'align-top':         ('vertical-align', 'top'),
            'align-middle':      ('vertical-align', 'middle'),
            'align-bottom':      ('vertical-align', 'bottom'),
            'align-text-top':    ('vertical-align', 'text-top'),
            'align-text-bottom': ('vertical-align', 'text-bottom'),
            'align-sub':         ('vertical-align', 'sub'),
            'align-super':       ('vertical-align', 'super'),
        }

        if self.joined_args(values):
            return True
        value = merge_value(self.args[1:])
        self.add_style('vertical-align', value)
        return True

    def whitespace(self):
        values = {
            'whitespace-normal':       ('white-space','normal'),
            'whitespace-nowrap':       ('white-space','nowrap'),
            'whitespace-pre':          ('white-space','pre'),
            'whitespace-pre-line':     ('white-space','pre-line'),
            'whitespace-pre-wrap':     ('white-space','pre-wrap'),
            'whitespace-break-spaces': ('white-space','break-spaces'),
        }
        return self.joined_args(values)

    def hyphens(self):
        values = {
            'hyphens-none':   ('hyphens','none'),
            'hyphens-manual': ('hyphens','manual'),
            'hyphens-auto':   ('hyphens','auto'),
        }
        return self.joined_args(values)

    def bg(self):
        # background blend mode
        values = {
            'bg-blend-normal': ('background-blend-mode','normal'),
            'bg-blend-multiply': ('background-blend-mode','multiply'),
            'bg-blend-screen': ('background-blend-mode','screen'),
            'bg-blend-overlay': ('background-blend-mode','overlay'),
            'bg-blend-darken': ('background-blend-mode','darken'),
            'bg-blend-lighten': ('background-blend-mode','lighten'),
            'bg-blend-color-dodge': ('background-blend-mode','color-dodge'),
            'bg-blend-color-burn': ('background-blend-mode','color-burn'),
            'bg-blend-hard-light': ('background-blend-mode','hard-light'),
            'bg-blend-soft-light': ('background-blend-mode','soft-light'),
            'bg-blend-difference': ('background-blend-mode','difference'),
            'bg-blend-exclusion': ('background-blend-mode','exclusion'),
            'bg-blend-hue': ('background-blend-mode','hue'),
            'bg-blend-saturation': ('background-blend-mode','saturation'),
            'bg-blend-color': ('background-blend-mode','color'),
            'bg-blend-luminosity': ('background-blend-mode','luminosity'),
        }
        if self.joined_args(values):
            return True

        # background attachment: how a background image hehaves when scrolling
        values = {
            'bg-fixed':  ('background-attachment','fixed'),
            'bg-local':  ('background-attachment','local'),
            'bg-scroll': ('background-attachment','scroll'),
        }
        if self.joined_args(values):
            return True

        # background clip
        values = {
            'bg-clip-border':  ('background-clip','border-box'),
            'bg-clip-padding': ('background-clip','padding-box'),
            'bg-clip-content': ('background-clip','content-box'),
            'bg-clip-text':    ('background-clip','text'),
        }
        if self.joined_args(values):
            return True

        # background color
        color = convert_color(self.args[1:])
        if color:
            self.add_style('background-color', color)
            return True

        # background origin
        values = {
            'bg-origin-border':  ('background-origin','border-box'),
            'bg-origin-padding': ('background-origin','padding-box'),
            'bg-origin-content': ('background-origin','content-box'),
        }
        if self.joined_args(values):
            return True

        # background position
        values = {
            'bg-bottom':       ('background-position','bottom'),
            'bg-center':       ('background-position','center'),
            'bg-left':         ('background-position','left'),
            'bg-left-bottom':  ('background-position','left bottom'),
            'bg-left-top':     ('background-position','left top'),
            'bg-right':        ('background-position','right'),
            'bg-right-bottom': ('background-position','right bottom'),
            'bg-right-top':    ('background-position','right top'),
            'bg-top':          ('background-position','top'),
        }
        if self.joined_args(values):
            return True
        if self.argc > 2 and self.args[1] == 'pos':
            value = merge_value(self.args[2:])
            self.add_style('background-position', value)
            return True

        # background repeat
        values = {
            'bg-repeat':       ('background-repeat','repeat'),
            'bg-no-repeat':    ('background-repeat','no-repeat'),
            'bg-repeat-x':     ('background-repeat','repeat-x'),
            'bg-repeat-y':     ('background-repeat','repeat-y'),
            'bg-repeat-round': ('background-repeat','round'),
            'bg-repeat-space': ('background-repeat','space'),
        }
        if self.joined_args(values):
            return True
            
        # background size
        values = {
            'bg-auto':    ('background-size','auto'),
            'bg-cover':   ('background-size','cover'),
            'bg-contain': ('background-size','contain'),
        }
        if self.joined_args(values):
            return True
        if self.argc > 2 and self.args[1] == 'size':
            value = merge_value(self.args[2:])
            self.add_style('background-size', value)
            return True

        # background image
        values = {
            'bg-none':           ('background-image','none'),
            'bg-gradient-to-t':  ('background-image','linear-gradient(to top, var(--fry-gradient-stops))'),
            'bg-gradient-to-tr': ('background-image','linear-gradient(to top right, var(--fry-gradient-stops))'),
            'bg-gradient-to-r':  ('background-image','linear-gradient(to right, var(--fry-gradient-stops))'),
            'bg-gradient-to-br': ('background-image','linear-gradient(to bottom right, var(--fry-gradient-stops))'),
            'bg-gradient-to-b':  ('background-image','linear-gradient(to bottom, var(--fry-gradient-stops))'),
            'bg-gradient-to-bl': ('background-image','linear-gradient(to bottom left, var(--fry-gradient-stops))'),
            'bg-gradient-to-l':  ('background-image','linear-gradient(to left, var(--fry-gradient-stops))'),
            'bg-gradient-to-tl': ('background-image','linear-gradient(to top left, var(--fry-gradient-stops))'),
            'bg-radial':         ('background-image','radial-gradient(var(--fry-gradient-stops))'),
        }
        if self.joined_args(values):
            return True
        if self.argc > 2 and self.args[1] == 'img':
            value = merge_value(self.args[2:])
            self.add_style('background-image', value)
            return True

    def from_method(self):
        if self.argc == 1:
            return False

        # from position
        if self.argc == 2 and is_digit_range(self.args[1], 0, 100):
            self.add_style('--fry-gradient-from-position', f'{self.args[1]}%')
            return True
        if self.argc == 2 and is_percent(self.args[1]):
            self.add_style('--fry-gradient-from-position', self.args[1])
            return True


        # from color
        color = convert_color(self.args[1:])
        if not color:
            return False
        opacity0 = opacity_to(color, 0)
        self.add_style('--fry-gradient-from', f'{color} var(--fry-gradient-from-position)')
        self.add_style('--fry-gradient-to', f'{opacity0} var(--fry-gradient-to-position)')
        self.add_style('--fry-gradient-stops', 'var(--fry-gradient-from), var(--fry-gradient-to)')
        return True


    def via(self):
        if self.argc == 1:
            return False

        # via position
        if self.argc == 2 and is_digit_range(self.args[1], 0, 100):
            self.add_style('--fry-gradient-via-position', f'{self.args[1]}%')
            return True
        if self.argc == 2 and is_percent(self.args[1]):
            self.add_style('--fry-gradient-via-position', self.args[1])
            return True

        # via color
        color = convert_color(self.args[1:])
        if not color:
            return False
        opacity0 = opacity_to(color, 0)
        self.add_style('--fry-gradient-to', f'{opacity0} var(--fry-gradient-to-position)')
        self.add_style('--fry-gradient-stops', f'var(--fry-gradient-from), {color} var(--fry-gradient-via-position), var(--fry-gradient-to)')
        return True

    def to(self):
        if self.argc == 1:
            return False

        # to position
        if self.argc == 2 and is_digit_range(self.args[1], 0, 100):
            self.add_style('--fry-gradient-to-position', f'{self.args[1]}%')
            return True
        if self.argc == 2 and is_percent(self.args[1]):
            self.add_style('--fry-gradient-to-position', self.args[1])
            return True

        # to color
        color = convert_color(self.args[1:])
        if not color:
            return False
        self.add_style('--fry-gradient-to', f'{color} var(--fry-gradient-to-position)')
        return True

    def rounded(self):
        default_size = '0.25rem'
        sizes = {
            'none': '0px',
            'sm':   '0.125rem',
            'md':   '0.375rem',
            'lg':   '0.5rem',
            'xl':   '0.75rem',
            '2xl':  '1rem',
            '3xl':  '1.5rem',
            'full': '9999px',
        }
        positions = {
            's':  ('start-start', 'end-start'),
            'e':  ('start-end', 'end-end'),
            't':  ('top-left', 'top-right'),
            'r':  ('top-right', 'bottom-right'),
            'b':  ('bottom-right', 'bottom-left'),
            'l':  ('top-left', 'bottom-left'),
            'ss': ('start-start',),
            'se': ('start-end',),
            'ee': ('end-end',),
            'es': ('end-start',),
            'tl': ('top-left',),
            'tr': ('top-right',),
            'bl': ('bottom-left',),
            'br': ('bottom-right',),
        }
        if self.argc == 1:
            self.add_style('border-radius', default_size)
            return True
        if self.argc == 2:
            if self.args[1] in sizes:
                self.add_style('border-radius', sizes[self.args[1]])
                return True
            if self.args[1] in positions:
                for pos in positions[self.args[1]]:
                    self.add_style(f'border-{pos}-radius', default_size)
                return True
            value = merge_value(self.args[1:])
            self.add_style('border-radius', value)
            return True
        if self.args[1] not in positions:
            return False
        if self.argc == 3 and self.args[2] in sizes:
            size = sizes[self.args[2]]
        else:
            size = merge_value(self.args[2:])
        for pos in positions[self.args[1]]:
            self.add_style(f'border-{pos}-radius', size)
        return True

    def border(self):
        # set table border collapse
        values = {
            'border-collapse': ('border-collapse','collapse'),
            'border-separate': ('border-collapse','separate'),
        }

        if self.joined_args(values):
            return True

        # set border style
        values = {
            'border-solid':  ('border-style','solid'),
            'border-dashed': ('border-style','dashed'),
            'border-dotted': ('border-style','dotted'),
            'border-double': ('border-style','double'),
            'border-hidden': ('border-style','hidden'),
            'border-none':   ('border-style','none'),
        }

        if self.joined_args(values):
            return True

        # set table border spacing
        if self.argc > 2 and self.args[1] == 'spacing':
            if self.args[2] in 'xy' and self.argc == 4:
                size = convert_size(self.args[3])
                self.add_style(f'--fry-border-spacing-{self.args[2]}', size)
            elif self.argc == 3:
                size = convert_size(self.args[2])
                self.add_style('--fry-border-spacing-x', size)
                self.add_style('--fry-border-spacing-y', size)
            else:
                return False
            self.add_style('border-spacing', 'var(--fry-border-spacing-x) var(--fry-border-spacing-y)')
            return True

        # set border width and border color
        positions = {
            'x': ('left', 'right'),
            'y': ('top', 'bottom'),
            's': ('inline-start',),
            'e': ('inline-end',),
            't': ('top',),
            'r': ('right',),
            'b': ('bottom',),
            'l': ('left',),
        }
        default_size = 1

        # border设置宽度为1px
        if self.argc == 1:
            self.add_style('border-width', f'{default_size}px')
            return True
        color = convert_color(self.args[1:])
        if color:
            self.add_style('border-color', color)
            return True

        if self.argc == 2:
            if self.args[1] in positions:
                for pos in positions[self.args[1]]:
                    self.add_style(f'border-{pos}-width', f'{default_size}px')
                return True
            if self.args[1].isdigit():
                self.add_style('border-width', f'{self.args[1]}px')
                return True
            value = merge_value(self.args[1:])
            self.add_style('border-width', value)
            return True
        if self.args[1] not in positions:
            return False

        color = convert_color(self.args[2:])
        if color:
            for pos in positions[self.args[1]]:
                self.add_style(f'border-{pos}-color', color)
            return True

        if self.argc == 3 and self.args[2].isdigit():
            size = f'{self.args[2]}px'
        else:
            size = merge_value(self.args[2:])
        for pos in positions[self.args[1]]:
            self.add_style(f'border-{pos}-width', size)
        return True

    def divide(self):
        # divide作用于其直接子元素
        self.css.selector_template += ' > :not([hidden]) ~ :not([hidden])'

        # divide styles
        values = {
            'divide-solid':  ('border-style','solid'),
            'divide-dashed': ('border-style','dashed'),
            'divide-dotted': ('border-style','dotted'),
            'divide-double': ('border-style','double'),
            'divide-none':   ('border-style','none'),
        }

        if self.joined_args(values):
            return True

        if self.argc == 1:
            return False

        # divide color
        color = convert_color(self.args[1:])
        if color:
            self.add_style('border-color', color)
            return True

        # divide width
        if self.args[1] not in 'xy':
            return False
        if self.argc == 2:
            width = '1px'
        elif self.args[2].isdigit():
            width = f'{self.args[2]}px'
        elif self.args[2] == 'reverse':
            self.add_style(f'--fry-divide-{self.args[1]}-reverse', '1')
            return True
        else:
            width = merge_value(self.args[2:])
        if self.args[1] == 'x':
            self.add_style('--fry-divide-x-reverse', '0')
            self.add_style('border-left-width', f'calc({width} * calc(1 - var(--fry-divide-x-reverse)))')
            self.add_style('border-right-width', f'calc({width} * var(--fry-divide-x-reverse))')
        else:
            self.add_style('--fry-divide-y-reverse', '0')
            self.add_style('border-top-width', f'calc({width} * calc(1 - var(--fry-divide-y-reverse)))')
            self.add_style('border-bottom-width', f'calc({width} * var(--fry-divide-y-reverse))')
        return True

    def outline(self):
        # outline style
        values = {
            'outline-none':   [('outline', '2px solid transparent'), ('outline-offset', '2px')],
            'outline':        ('outline-style', 'solid'),
            'outline-dashed': ('outline-style', 'dashed'),
            'outline-dotted': ('outline-style', 'dotted'),
            'outline-double': ('outline-style', 'double'),
        }
        if self.joined_args(values):
            return True

        # outline color
        color = convert_color(self.args[1:])
        if color:
            self.add_style('outline-color', color)
            return True
        
        # outline offset
        if self.args[1] == 'offset' and self.argc > 2:
            if self.argc == 3 and self.args[2].isdigit():
                offset = f'{self.args[2]}px'
            else:
                offset = merge_value(self.args[2:])
            self.add_style('outline-offset', offset)
            return True

        # outline width
        if self.argc == 2 and self.args[1].isdigit():
            width = f'{self.args[1]}px'
        else:
            width = merge_value(self.args[1:])
        self.add_style('outline-width', width)
        return True

    def ring(self):
        # ring inset
        values = {
            'ring-inset': ('--fry-ring-inset','inset'),
        }

        if self.joined_args(values):
            return True

        # ring color
        color = convert_color(self.args[1:])
        if color:
            self.add_style('--fry-ring-color', color)
            return True

        # ring offset width and ring offset color
        if self.argc > 2 and self.args[1] == 'offset':
            color = convert_color(self.args[2:])
            if color:
                self.add_style('--fry-ring-offset-color', color)
                return True
            if self.args[2].isdigit():
                width = f'{self.args[2]}px'
            else:
                width = merge_value(self.args[2:])
            self.add_style('--fry-ring-offset-width', width)
            return True

        # ring width
        if self.argc == 1:
            width = '3px'
        elif self.argc == 2 and self.args[1].isdigit():
            width = f'{self.args[1]}px'
        else:
            width = merge_value(self.args[1:])
        self.add_style('--fry-ring-offset-shadow', 'var(--fry-ring-inset) 0 0 0 var(--fry-ring-offset-width) var(--fry-ring-offset-color)')
        self.add_style('--fry-ring-shadow', f'var(--fry-ring-inset) 0 0 0 calc({width} + var(--fry-ring-offset-width)) var(--fry-ring-color)')
        self.add_style('box-shadow', 'var(--fry-ring-offset-shadow), var(--fry-ring-shadow), var(--fry-shadow, 0 0 #0000)')
        return True

    def shadow(self):
        values = {
            'sm':     ('0 1px 2px 0 rgb(0 0 0 / 0.05)',
                       '0 1px 2px 0 var(--fry-shadow-color)'),
            'nm':     ('0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1)',
                       '0 1px 3px 0 var(--fry-shadow-color), 0 1px 2px -1px var(--fry-shadow-color)'),
            'md':     ('0 4px 6px -1px rgb(0 0 0 / 0.1), 0 2px 4px -2px rgb(0 0 0 / 0.1)',
                       '0 4px 6px -1px var(--fry-shadow-color), 0 2px 4px -2px var(--fry-shadow-color)'),
            'lg':     ('0 10px 15px -3px rgb(0 0 0 / 0.1), 0 4px 6px -4px rgb(0 0 0 / 0.1)',
                       '0 10px 15px -3px var(--fry-shadow-color), 0 4px 6px -4px var(--fry-shadow-color)'),
            'xl':     ('0 20px 25px -5px rgb(0 0 0 / 0.1), 0 8px 10px -6px rgb(0 0 0 / 0.1)',
                       '0 20px 25px -5px var(--fry-shadow-color), 0 8px 10px -6px var(--fry-shadow-color)'),
            '2xl':    ('0 25px 50px -12px rgb(0 0 0 / 0.25)',
                       '0 25px 50px -12px var(--fry-shadow-color)'),
            'inner':  ('inset 0 2px 4px 0 rgb(0 0 0 / 0.05)',
                       'inset 0 2px 4px 0 var(--fry-shadow-color)'),
            'none':   ('0 0 #0000',
                       '0 0 #0000'),
        }

        # box shadow color
        color = convert_color(self.args[1:])
        if color:
            self.add_style('--fry-shadow-color', color)
            self.add_style('--fry-shadow', 'var(--fry-shadow-colored)')
            return True

        if self.argc == 1:
            size = 'nm'
        elif self.argc == 2 and self.args[1] in values:
            size = self.args[1]
        else:
            value = merge_value(self.args[1:])
            self.add_style('box-shadow', value)
            return True
        shadow1, shadow2 = values[size]
        self.add_style('--fry-shadow', shadow1)
        self.add_style('--fry-shadow-colored', shadow2)
        self.add_style('box-shadow', 'var(--fry-ring-offset-shadow, 0 0 #0000), var(--fry-ring-shadow, 0 0 #0000), var(--fry-shadow)')
        return True

    def filter_opacity(self, argc, args):
        if argc != 2:
            return None
        if not args[1].isdigit():
            return None
        value = int(self.args[1])
        if value < 0 or value > 100:
            return None
        return value/100

    def opacity(self):
        value = self.filter_opacity(self.argc, self.args)
        if value is None:
            return False
        self.add_style('opacity', str(value))
        return True

    def mix(self):
        values = {
            'mix-blend-normal':       ('mix-blend-mode','normal'),
            'mix-blend-multiply':     ('mix-blend-mode','multiply'),
            'mix-blend-screen':       ('mix-blend-mode','screen'),
            'mix-blend-overlay':      ('mix-blend-mode','overlay'),
            'mix-blend-darken':       ('mix-blend-mode','darken'),
            'mix-blend-lighten':      ('mix-blend-mode','lighten'),
            'mix-blend-color-dodge':  ('mix-blend-mode','color-dodge'),
            'mix-blend-color-burn':   ('mix-blend-mode','color-burn'),
            'mix-blend-hard-light':   ('mix-blend-mode','hard-light'),
            'mix-blend-soft-light':   ('mix-blend-mode','soft-light'),
            'mix-blend-difference':   ('mix-blend-mode','difference'),
            'mix-blend-exclusion':    ('mix-blend-mode','exclusion'),
            'mix-blend-hue':          ('mix-blend-mode','hue'),
            'mix-blend-saturation':   ('mix-blend-mode','saturation'),
            'mix-blend-color':        ('mix-blend-mode','color'),
            'mix-blend-luminosity':   ('mix-blend-mode','luminosity'),
            'mix-blend-plus-lighter': ('mix-blend-mode','plus-lighter'),
        }

        return self.joined_args(values)

    filter_vars = 'var(--fry-blur) var(--fry-brightness) var(--fry-contrast) var(--fry-grayscale) var(--fry-hue-rotate) var(--fry-invert) var(--fry-saturate) var(--fry-sepia) var(--fry-drop-shadow)'

    backdrop_filter_vars = 'var(--fry-backdrop-blur) var(--fry-backdrop-brightness) var(--fry-backdrop-contrast) var(--fry-backdrop-grayscale) var(--fry-backdrop-hue-rotate) var(--fry-backdrop-invert) var(--fry-backdrop-opacity) var(--fry-backdrop-saturate) var(--fry-backdrop-sepia)'

    def filter_blur(self, argc, args):
        sizes = {
            'none': '0',
            'sm':   '4px',
            'nm':   '8px',
            'md':   '12px',
            'lg':   '16px',
            'xl':   '24px',
            '2xl':  '40px',
            '3xl':  '64px',
        }

        if argc == 1:
            size = sizes['nm']
        elif argc == 2 and args[1] in sizes:
            size = sizes[args[1]]
        else:
            size = merge_value(args[1:])
        return size

    def blur(self):
        size = self.filter_blur(self.argc, self.args)
        self.add_style('--fry-blur', f'blur({size})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter_brightness(self, argc, args):
        if argc != 2 or not args[1].isdigit():
            return None
        value = int(args[1])
        if value < 0 or value > 200:
            return None
        return value/100

    def brightness(self):
        value = self.filter_brightness(self.argc, self.args)
        if value is None:
            return False
        self.add_style('--fry-brightness', f'brightness({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter_contrast(self, argc, args):
        if argc != 2 or not args[1].isdigit():
            return None
        value = int(args[1])
        if value < 0 or value > 200:
            return None
        return value/100

    def contrast(self):
        value = self.filter_contrast(self.argc, self.args)
        if value is None:
            return False
        self.add_style('--fry-contrast', f'contrast({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def drop(self):
        if self.argc < 2 or self.args[1] != 'shadow':
            return False

        values = {
            'sm':   ('0 1px 1px rgb(0 0 0 / 0.05)',),
            'nm':   ('0 1px 2px rgb(0 0 0 / 0.1))', '0 1px 1px rgb(0 0 0 / 0.06)'),
            'md':   ('0 4px 3px rgb(0 0 0 / 0.07)', '0 2px 2px rgb(0 0 0 / 0.06)'),
            'lg':   ('0 10px 8px rgb(0 0 0 / 0.04)', '0 4px 3px rgb(0 0 0 / 0.1)'),
            'xl':   ('0 20px 13px rgb(0 0 0 / 0.03)', '0 8px 5px rgb(0 0 0 / 0.08)'),
            '2xl':  ('0 25px 25px rgb(0 0 0 / 0.15)',),
            'none': ('0 0 #0000',),
        }
        if self.argc == 2:
            value = values['nm']
        elif self.argc == 3 and self.args[2] in values:
            value = values[self.args[2]]
        elif self.argc == 3:
            value = (merge_value(self.args[2:]),)
        elif self.argc == 4:
            value = (merge_value([self.args[2]]), merge_value([self.args[3]]))
        else:
            return False
        self.add_style('--fry-drop-shadow', ' '.join(f'drop-shadow({v})' for v in value))
        self.add_style('filter', self.filter_vars)
        return True

    def filter_grayscale(self, argc, args):
        if argc == 1:
            value = 100
        elif argc == 2 and args[1].isdigit():
            value = int(args[1])
            if value < 0 or value > 100:
                return None
        else:
            return None
        return f'{value}%'

    def grayscale(self):
        value = self.filter_grayscale(self.argc, self.args)
        if value is None:
            return False
        self.add_style('--fry-grayscale', f'grayscale({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter_hue(self, argc, args, neg):
        if argc != 3 and args[1] != 'rotate' and not args[2].isdigit():
            return None
        value = int(args[2])
        if value < 0 or value > 360:
            return None
        if neg:
            value = f'-{value}deg'
        else:
            value = f'{value}deg'
        return value

    def hue(self):
        value = self.filter_hue(self.argc, self.args, self.neg)
        if value is None:
            return False
        self.add_style('--fry-hue-rotate', f'hue-rotate({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter_invert(self, argc, args):
        if argc == 1:
            value = 100
        elif argc == 2 and args[1].isdigit():
            value = int(args[1])
            if value < 0 or value > 100:
                return None
        else:
            return None
        return f'{value}%'

    def invert(self):
        value = self.filter_invert(self.argc, self.args)
        if value is None:
            return False
        self.add_style('--fry-invert', f'invert({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter_saturate(self, argc, args):
        if argc != 2 or not args[1].isdigit():
            return None
        value = int(args[1])
        if value < 0 or value > 200:
            return None
        return value/100

    def saturate(self):
        value = self.filter_saturate(self.argc, self.args)
        if value is None:
            return False
        self.add_style('--fry-saturate', f'saturate({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter_sepia(self, argc, args):
        if argc == 1:
            value = 100
        elif argc == 2 and args[1].isdigit():
            value = int(args[1])
            if value < 0 or value > 100:
                return None
        else:
            return None
        return f'{value}%'

    def sepia(self):
        value = self.filter_sepia(self.argc, self.args)
        if value is None:
            return False
        self.add_style('--fry-sepia', f'sepia({value})')
        self.add_style('filter', self.filter_vars)
        return True

    def filter(self):
        values = {
            'filter-none': ('filter','none'),
        }
        return self.joined_args(values)

    def backdrop(self):
        values = {
            'backdrop-filter-none': [('-webkit-backdrop-filter', 'none'),
                                             ('backdrop-filter', 'none')],
        }
        if self.joined_args(values):
            return True

        if self.argc == 1:
            return False
        
        argc = self.argc - 1
        args = self.args[1:]
        neg  = self.neg

        if self.args[1] == 'blur':
            size = self.filter_blur(argc, args)
            self.add_style('--fry-backfrop-blur', f'blur({size})')
        elif self.args[1] == 'brightness':
            value = self.filter_brightness(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-brightness', f'brightness({value})')
        elif self.args[1] == 'contrast':
            value = self.filter_contrast(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-contrast', f'contrast({value})')
        elif self.args[1] == 'grayscale':
            value = self.filter_grayscale(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-grayscale', f'grayscale({value})')
        elif self.args[1] == 'hue':
            value = self.filter_hue(argc, args, neg)
            if value is None:
                return False
            self.add_style('--fry-backdrop-hue-rotate', f'hue-rotate({value})')
        elif self.args[1] == 'invert':
            value = self.filter_invert(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-invert', f'invert({value})')
        elif self.args[1] == 'opacity':
            value = self.filter_opacity(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-opacity', f'opacity({value})')
        elif self.args[1] == 'saturate':
            value = self.filter_saturate(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-saturate', f'saturate({value})')
        elif self.args[1] == 'sepia':
            value = self.filter_sepia(argc, args)
            if value is None:
                return False
            self.add_style('--fry-backdrop-sepia', f'sepia({value})')
        else:
            return False

        self.add_style('-webkit-backdrop-filter', self.backdrop_filter_vars)
        self.add_style('backdrop-filter', self.backdrop_filter_vars)
        return True

    def caption(self):
        # set table caption side
        values = {
            'caption-top':    ('caption-side','top'),
            'caption-bottom': ('caption-side','bottom'),
        }
        return self.joined_args(values)

    def transition(self):
        # transition none
        values = {
            'transition-none': ('transition-property','none'),
        }

        if self.joined_args(values):
            return True

        # transition properties
        values = {
            'transation-all': ('transition-property','all'),
            'transition':     ('transition-property','color, background-color, border-color, text-decoration-color, fill, stroke, opacity, box-shadow, transform, filter, backdrop-filter'),
            'transition-colors': ('transition-property','color, background-color, border-color, text-decoration-color, fill, stroke'),
            'transition-opacity': ('transition-property','opacity'),
            'transition-shadow': ('transition-property','box-shadow'),
            'transition-transform': ('transition-property','transform'),
        }

        if not self.joined_args(values):
            prop = merge_value(self.args[1:])
            self.add_style('transition-property', prop)
        self.add_style('transition-timing-function', 'cubic-bezier(0.4, 0, 0.2, 1)')
        self.add_style('transition-duration', '150ms')
        return True
    
    def duration(self):
        if self.argc != 2 and not self.args[1].isdigit():
            return False
        self.add_style('transition-duration', f'{self.args[1]}ms')
        return True
    
    def ease(self):
        if self.argc == 1:
            return False
        values = {
            'ease-linear': ('transition-timing-function','linear'),
            'ease-in':     ('transition-timing-function','cubic-bezier(0.4, 0, 1, 1)'),
            'ease-out':    ('transition-timing-function','cubic-bezier(0, 0, 0.2, 1)'),
            'ease-in-out': ('transition-timing-function','cubic-bezier(0.4, 0, 0.2, 1)'),
        }
        if not self.joined_args(values):
            timing = merge_value(self.args[1])
            self.add_style('transition-timing-function', timing)
        return True

    def delay(self):
        if self.argc != 2 and not self.args[1].isdigit():
            return False
        self.add_style('transition-delay', f'{self.args[1]}ms')
        return True
        
    def animate(self):
        if self.argc == 1:
            return False
        values = {
            'animate-none':              ('animation','none'),
            'animate-spin':              ('animation','spin 1s linear infinite'),
            'animate-ping':              ('animation','ping 1s cubic-bezier(0, 0, 0.2, 1) infinite'),
            'animate-pulse':             ('animation','pulse 2s cubic-bezier(0.4, 0, 0.6, 1) infinite'),
            'animate-bounce':            ('animation','bounce 1s infinite'),
            'animate-normal':            ('animation-direction', 'normal'),
            'animate-reverse':           ('animation-direction', 'reverse'),
            'animate-alternate':         ('animation-direction', 'alternate'),
            'animate-alternate-reverse': ('animation-direction', 'alternate-reverse'),
            'animate-forwards':          ('animation-fill-mode', 'forwards'),
            'animate-backwards':         ('animation-fill-mode', 'backwards'),
            'animate-infinite':          ('animation-iteration-count', 'infinite'),
            'animate-paused':            ('animation-play-state', 'paused'),
            'animate-running':           ('animation-play-state', 'running'),
            'animate-ease':              ('animation-timing-function', 'ease'),
            'animate-ease-in':           ('animation-timing-function', 'ease-in'),
            'animate-ease-out':          ('animation-timing-function', 'ease-out'),
            'animate-ease-in-out':       ('animation-timing-function', 'ease-in-out'),
            'animate-linear':            ('animation-timing-function', 'linear'),
        }

        if self.joined_args(values):
            return True

        if self.argc > 2:
            # 支持多个animation的情况，需要逗号隔开
            sub = self.args[1]
            if sub == 'name':
                key = 'animation-name'
            elif sub == 'duration':
                key = 'animation-duration'
            elif sub == 'delay':
                key = 'animation-delay'
            elif sub == 'dir':
                key = 'animation-direction'
            elif sub == 'fill':
                key = 'animation-fill-mode'
            elif sub == 'count':
                key = 'animation-iteration-count'
            elif sub == 'state':
                key = 'animation-play-state'
            elif sub == 'ease':
                key = 'animation-timing-function'
                if self.argc == 6 and all(is_digit(v) or is_float(v) for v in self.args[2:]):
                    args = self.args[2:]
                    self.add_style(key, f'cubic-bezier({args[0]}, {args[1]}, {args[2]}, {args[3]})')
                    return True
            else:
                animation = merge_value(self.args[1:])
                self.add_style('animation', animation)
                return True
            value = merge_value(self.args[2:], self.neg)
            self.add_style(key, value)
        else:
            animation = merge_value(self.args[1:])
            self.add_style('animation', animation)
        return True
    
    transform_vars = 'translate(var(--fry-translate-x), var(--fry-translate-y)) rotate(var(--fry-rotate)) skewX(var(--fry-skew-x)) skewY(var(--fry-skew-y)) scaleX(var(--fry-scale-x)) scaleY(var(--fry-scale-y))'

    gpu_transform_vars = 'translate3d(var(--fry-translate-x), var(--fry-translate-y), 0) rotate(var(--fry-rotate)) skewX(var(--fry-skew-x)) skewY(var(--fry-skew-y)) scaleX(var(--fry-scale-x)) scaleY(var(--fry-scale-y))'

    def scale(self):
        if self.argc == 2 and self.args[1].isdigit():
            scale = self.args[1]
            dir = 'xy'
        elif self.argc == 3 and self.args[1] in 'xy' and self.args[2].isdigit():
            scale = self.args[2]
            dir = self.args[1]
        else:
            return False
        scale = int(self.args[1])/100
        scale = -scale if self.neg else scale
        if 'x' in dir:
            self.add_style('--fry-scale-x', str(scale))
        if 'y' in dir:
            self.add_style('--fry-scale-y', str(scale))
        self.add_style('transform', self.transform_vars)
        return True

    def rotate(self):
        if self.argc == 2 and self.args[1].isdigit():
            deg = ('-' if self.neg else '') + self.args[1]
            self.add_style('--fry-rotate', f'{deg}deg')
            self.add_style('transform', self.transform_vars)
            return True
        return False

    def translate(self):
        if self.argc == 3 and self.args[1] in 'xy':
            size = convert_size(self.args[2], self.neg)
            self.add_style(f'--fry-translate-{self.args[1]}', size)
            self.add_style('transform', self.transform_vars)
            return True
        return False

    def skew(self):
        if self.argc == 3 and self.args[1] in 'xy' and self.args[2].isdigit():
            deg = ('-' if self.neg else '') + self.args[2]
            self.add_style(f'--fry-skew-{self.args[1]}', f'{deg}deg')
            self.add_style('transform', self.transform_vars)
            return True
        return False

    def origin(self):
        if self.argc == 1:
            return False
        values = {
            'origin-center':       ('transform-origin','center'),
            'origin-top':          ('transform-origin','top'),
            'origin-top-right':    ('transform-origin','top right'),
            'origin-right':        ('transform-origin','right'),
            'origin-bottom-right': ('transform-origin','bottom right'),
            'origin-bottom':       ('transform-origin','bottom'),
            'origin-bottom-left':  ('transform-origin','bottom left'),
            'origin-left':         ('transform-origin','left'),
            'origin-top-left':     ('transform-origin','top left'),
        }
        if not self.joined_args(values):
            value = merge_value(self.args[1:])
            self.add_style('transform-origin', value)
        return True

    def transform(self):
        values = {
            'transform-none': ('transform','none'),
            'transform-gpu':  ('transform', self.gpu_transform_vars),
        }
        if self.joined_args(values):
            return True

    def accent(self):
        color = convert_color(self.args[1:])
        if color:
            self.add_style('accent-color', color)
            return True
        return False

    def appearance(self):
        values = {
            'appearance-none': ('appearance','none'),
        }
        return self.joined_args(values)

    def cursor(self):
        if self.argc == 1:
            return False
        value = merge_value(self.args[1:])
        self.add_style('cursor', value)
        return True

    def caret(self):
        color = convert_color(self.args[1:])
        if color:
            self.add_style('caret-color', color)
            return True
        return False

    def pointer(self):
        values = {
            'pointer-events-none': ('pointer-events','none'),
            'pointer-events-auto': ('pointer-events','auto'),
        }
        return self.joined_args(values)

    def resize(self):
        values = {
            'resize-none': ('resize','none'),
            'resize-y':    ('resize','vertical'),
            'resize-x':    ('resize','horizontal'),
            'resize':      ('resize','both'),
        }
        return self.joined_args(values)

    def scroll(self):
        values = {
            'scroll-auto':   ('scroll-behavior','auto'),
            'scroll-smooth': ('scroll-behavior','smooth'),
        }
        if self.joined_args(values):
            return True
        if self.argc == 3 and self.args[1][0] in 'mp':
            self.args = self.args[1:]
            self.argc = 2
            if self():
                styles = self.css.styles
                self.css.styles = [(f'scroll-{k}', v) for k, v in styles]
                return True
        return False

    def snap(self):
        values = {
            'snap-start':      ('scroll-snap-align','start'),
            'snap-end':        ('scroll-snap-align','end'),
            'snap-center':     ('scroll-snap-align','center'),
            'snap-align-none': ('scroll-snap-align','none'),
            'snap-normal':     ('scroll-snap-stop','normal'),
            'snap-always':     ('scroll-snap-stop','always'),
            'snap-none':       ('scroll-snap-type','none'),
            'snap-x':          ('scroll-snap-type','x var(--fry-scroll-snap-strictness)'),
            'snap-y':          ('scroll-snap-type','y var(--fry-scroll-snap-strictness)'),
            'snap-both':       ('scroll-snap-type','both var(--fry-scroll-snap-strictness)'),
            'snap-mandatory':  ('--fry-scroll-snap-strictness','mandatory'),
            'snap-proximity':  ('--fry-scroll-snap-strictness','proximity'),
        }
        return self.joined_args(values)

    def touch(self):
        values = {
            'touch-auto':         ('touch-action','auto'),
            'touch-none':         ('touch-action','none'),
            'touch-pan-x':        ('touch-action','pan-x'),
            'touch-pan-left':     ('touch-action','pan-left'),
            'touch-pan-right':    ('touch-action','pan-right'),
            'touch-pan-y':        ('touch-action','pan-y'),
            'touch-pan-up':       ('touch-action','pan-up'),
            'touch-pan-down':     ('touch-action','pan-down'),
            'touch-pinch-zoom':   ('touch-action','pinch-zoom'),
            'touch-manipulation': ('touch-action','manipulation'),
        }
        return self.joined_args(values)

    def select(self):
        values = {
            'select-none': ('user-select','none'),
            'select-text': ('user-select','text'),
            'select-all':  ('user-select','all'),
            'select-auto': ('user-select','auto'),
        }
        return self.joined_args(values)

    def will(self):
        values = {
            'will-change-auto':      ('will-change','auto'),
            'will-change-scroll':    ('will-change','scroll-position'),
            'will-change-contents':  ('will-change','contents'),
            'will-change-transform': ('will-change','transform'),
        }
        return self.joined_args(values)

    def fill(self):
        if self.argc == 1:
            return False
        color = convert_color(self.args[1:])
        if color:
            self.add_style('fill', color)
            return True
        return False

    def stroke(self):
        if self.argc == 1:
            return False
        color = convert_color(self.args[1:])
        if color:
            self.add_style('stroke', color)
        else:
            width = merge_value(self.args[1:])
            self.add_style('stroke-width', width)
        return True

    def sr(self):
        values = {
            'sr-only': [('position', 'absolute'),
                        ('width', '1px'),
                        ('height', '1px'),
                        ('padding', '0'),
                        ('margin', '-1px'),
                        ('overflow', 'hidden'),
                        ('clip', 'rect(0, 0, 0, 0)'),
                        ('white-space', 'nowrap'),
                        ('border-width', '0')],
        }
        return self.joined_args(values)




from parsimonious import NodeVisitor, BadGrammar
from collections import defaultdict

import re

from fryhcs.fileiter import FileIter
from fryhcs.fry.grammar import grammar
from fryhcs.spec import is_valid_html_attribute
from fryhcs.element import class_attr_name

class BaseCollector():
    ignored_tags = ('head', 'title', 'meta', 'style', 'link', 'script', 'template')

    def __init__(self):
        self.fileiter = FileIter()
        self.attrs = defaultdict(set)
        self.classes = set()

    def add_glob(self, path, glob):
        self.fileiter.add_glob(path, glob)

    def add_file(self, file):
        self.fileiter.add_file(file)

    def collect_attrs(self):
        for file in self.fileiter.all_files():
            # 设置newline=''确保在windows下换行符为\r\n，文件内容不会被open改变
            # 参考[universal newlines mode](https://docs.python.org/3/library/functions.html#open-newline-parameter)
            with file.open('r', encoding='utf-8', newline='') as f:
                self.collect_from_content(f.read())

    def collect_from_content(self, data):
        pass

    def collect_kv(self, k, v):
        if not v: v = ""
        if len(v) > 1 and v[0] in "\"'":
            v = v[1:-1]
        vs = v.split()
        if not k or k == 'class':
            self.classes.update(vs)
        else:
            self.attrs[k].update(vs)

    def all_attrs(self):
        for cls in self.classes:
            yield '', cls
        for k, vs in self.attrs.items():
            if not vs:
                vs.add('')
            for v in vs:
                yield k, v


class RegexCollector(BaseCollector):
    tagname = r'([a-zA-Z0-9]+)'
    attrname = r"""[^\s"'>/=]+"""
    attrvalue = r"""'[^']*'|"[^"]*"|[^\s"'=><`]+"""
    attr = r"""[^"'>]*|"[^"]*"|'[^']*'"""

    #(?:xxx)表示不取值的group
    starttag_re = re.compile(f"<{tagname}((?:{attr})*)/?>")
    attr_re = re.compile(f"({attrname})(?:\s*=\s*({attrvalue}))?")

    def collect_from_content(self, data):
        for starttag in self.starttag_re.finditer(data):
            name = starttag.group(1)
            attrs = starttag.group(2)
            if name in self.ignored_tags:
                continue
            if not attrs:
                continue
            for attr in self.attr_re.finditer(attrs):
                self.collect_kv(attr.group(1), attr.group(2))


class CssVisitor(NodeVisitor):
    def __init__(self, collect_kv):
        self.collect_kv = collect_kv

    def collect_literal(self, css_literal):
        if css_literal:
            for css in css_literal.split():
                eq = css.find('=')
                if eq >= 0:
                    key = css[:eq]
                    value = css[eq+1:]
                else:
                    key = css
                    value = ''
                self.collect_kv(key, value)

    def generic_visit(self, node, children):
        return None

    def visit_single_quote(self, node, children):
        return node.text

    def visit_double_quote(self, node, children):
        return node.text

    def visit_py_simple_quote(self, node, children):
        # python中的简单字符串常量加入收集范围
        key = children[0]
        if not ' ' in key:
            key = key[1:-1]
            self.collect_kv(key, '')

    def visit_js_simple_quote(self, node, children):
        # js中的简单字符串常量加入收集范围
        key = children[0]
        if not ' ' in key:
            key = key[1:-1]
            self.collect_kv(key, '')

    def visit_fry_self_closing_element(self, node, children):
        _, name, attrs, _, _ = children
        if name[0].islower():
            for attr in attrs:
                if not isinstance(attr, tuple):
                    print(f'attr type {type(attr)}: {attr}')
                    raise BadGrammar
                if is_valid_html_attribute(name, attr[0]):
                    continue
                self.collect_kv(attr[0], attr[1])
        else:
            for attr in attrs:
                if not isinstance(attr, tuple):
                    print(f'attr type {type(attr)}: {attr}')
                    raise BadGrammar
                if attr[0] == '':
                    self.collect_kv('', attr[1])

    def visit_fry_void_element(self, node, children):
        _, name, attrs, _, _ = children
        for attr in attrs:
            if not isinstance(attr, tuple):
                print(f'attr type {type(attr)}: {attr}')
                raise BadGrammar
            # void_element必定是html元素
            if is_valid_html_attribute(name, attr[0]):
                continue
            self.collect_kv(attr[0], attr[1])

    def visit_fry_start_tag(self, node, children):
        _, start_name, attrs, _, _ = children
        if start_name[0].islower():
            for attr in attrs:
                if not isinstance(attr, tuple):
                    raise BadGrammar
                if is_valid_html_attribute(start_name, attr[0]):
                    continue
                self.collect_kv(attr[0], attr[1])
        else:
            for attr in attrs:
                if not isinstance(attr, tuple):
                    raise BadGrammar
                if attr[0] == '':
                    self.collect_kv('', attr[1])

    def visit_fry_element_name(self, node, children):
        return node.text

    def visit_fry_void_element_name(self, node, children):
        return node.text

    def visit_fry_attributes(self, node, children):
        return [ch for ch in children if ch]

    def visit_fry_spaced_attribute(self, node, children):
        _, attr = children
        return attr

    def visit_fry_attribute(self, node, children):
        return children[0]

    def visit_fry_kv_attribute(self, node, children):
        name, _, _, _, value = children
        if name == class_attr_name:
            # 将class名字设置为空字符串，is_valid_html_attribute返回false
            # 否则不会收集相关utilities。
            name = ''
        if isinstance(value, str):
            return (name, value)

    def visit_fry_novalue_attribute(self, node, children):
        name, _ = children
        return (name, '')

    def visit_fry_attribute_name(self, node, children):
        return node.text

    def visit_fry_attribute_value(self, node, children):
        return children[0]

    def visit_single_f_string(self, node, children):
        return node.text

    def visit_double_f_string(self, node, children):
        return node.text

class ParserCollector(BaseCollector):
    def collect_from_content(self, data):
        tree = grammar.parse(data)
        visitor = CssVisitor(self.collect_kv)
        visitor.visit(tree)

Collector = ParserCollector


if __name__ == '__main__':
    collector = Collector()
    collector.add_glob('test', '**/*.html')
    collector.collect_attrs()
    print("classes:")
    for cls in collector.classes:
        print("", cls)

    print("attrs:")
    for k,v in collector.attrs.items():
        print("", k, "\t\t", v)



"""
扩展python import机制，直接import fry文件，在线转换成py文件，然后编译为.pyc文件直接执行。
"""
import sys
from importlib.machinery import FileFinder, SourceFileLoader
from importlib._bootstrap_external import _get_supported_file_loaders
from .generator import fry_to_py

# 最初.fry文件叫.pyx文件，参考.jsx，但.pyx在Cython中已经使用了，改为.fry文件
# 如果仍用.pyx，会在Cython使用时报错（如sqlalchemy/cyextension/collections.pyx）
PYXSOURCE_SUFFIXES = ['.fry',]

class PyxSourceFileLoader(SourceFileLoader):
    def source_to_code(self, data, path, *, _optimize=-1):
        data = fry_to_py(data.decode(), path)
        return super(SourceFileLoader, self).source_to_code(data, path, _optimize=_optimize)

def install_path_hook():
    if sys.path_hooks and hasattr(sys.path_hooks[0], 'fryhcs'):
        return
    loader_details = [(PyxSourceFileLoader, PYXSOURCE_SUFFIXES)] + _get_supported_file_loaders()
    factory_func = FileFinder.path_hook(*loader_details)
    setattr(factory_func, 'fryhcs', True)
    sys.path_hooks.insert(0, factory_func)

    # 清空已有的PathEntryFinder缓存
    sys.path_importer_cache.clear()



from pygments.lexer import Lexer, bygroups, default, include, inherit
from pygments.lexers.python import PythonLexer
from pygments.lexers.javascript import JavascriptLexer
from pygments.token import Token, Name, Operator, Punctuation, String, Text, Whitespace, Comment
from parsimonious import NodeVisitor, BadGrammar
from fryhcs.fry.grammar import grammar


def merge(children):
    result = []
    for child in children:
        if isinstance(child, list):
            new = merge(child)
            if result and new and result[-1][0] == new[0][0]:
                result[-1] = (new[0][0], result[-1][1]+new[0][1])
                new = new[1:]
            result += new
        elif isinstance(child, tuple):
            if len(child[1]) == 0:
                continue
            if result and result[-1][0] == child[0]:
                result[-1] = (child[0], result[-1][1]+child[1])
            else:
                result.append(child)
        else:
            raise BadGrammar(f"invalid: |{child}|, {type(child)}")
    return result

def an(text):
    return (Name.Attribute, text)

def en(text):
    if text[0].islower():
        return (Name.HtmlElement, text)
    else:
        return (Name.ComponentElement, text)

def n(text):
    return (Name, text)

def o(text):
    return (Operator, text)

def p(text):
    return (Punctuation, text)

def ep(text):
    return (Punctuation.ElementPunctuation, text)

def sep(text):
    return (Punctuation.ServerEmbedPunctuation, text)

def cep(text):
    return (Punctuation.ClientEmbedPunctuation, text)

def s(text):
    return (String, text)

def w(text):
    return (Whitespace, text)

def t(text):
    return (Text.HtmlText, text)

def py(text):
    return ('python', text)

def js(text):
    return ('javascript', text)

def fs(text):
    return ('fstring', text)

class FryVisitor(NodeVisitor):
    def generic_visit(self, node, children):
        return children or node.text

    def visit_fry_script(self, node, children):
        return merge(children)

    def visit_inner_fry_script(self, node, children):
        return children

    def visit_fry_script_item(self, node, children):
        return children[0]

    def visit_inner_fry_script_item(self, node, children):
        return children[0]

    def visit_py_comment(self, node, children):
        return py(node.text)

    def visit_inner_fry_brace(self, node, children):
        _l, script, _r = children
        return [py('{'), script, py('}')]

    def visit_fry_embed(self, node, children):
        _l, script, _r = children
        return [sep('{'), script, sep('}')]
        
    def visit_triple_single_quote(self, node, children):
        return py(node.text)

    def visit_triple_double_quote(self, node, children):
        return py(node.text)

    def visit_single_quote(self, node, children):
        return node.text

    def visit_double_quote(self, node, children):
        return node.text

    def visit_py_simple_quote(self, node, children):
        return py(children[0])

    def visit_js_simple_quote(self, node, children):
        return js(children[0])

    def visit_less_than_char(self, node, children):
        return py('<')

    def visit_no_component_d_char(self, node, children):
        return py('d')

    def visit_py_normal_code(self, node, children):
        return py(node.text)

    def visit_inner_py_normal_code(self, node, children):
        return py(node.text)

    def visit_fry_component(self, node, children):
        return children

    def visit_fry_component_header(self, node, children):
        return py(node.text)

    def visit_fry_web_template(self, node, children):
        l, ws1, element, ws2, r = children
        return [en(l), w(ws1), element, w(ws2), en(r)]

    def visit_fry_root_element(self, node, children):
        return children[0]

    def visit_fry_element(self, node, children):
        return children[0]

    def visit_fry_fragment(self, node, children):
        l, chs, r = children
        return [ep(l), chs, ep(r)]

    def visit_fry_self_closing_element(self, node, children):
        l, name, attrs, s, r = children
        return [ep(l), name, attrs, w(s), ep(r)]

    def visit_fry_void_element(self, node, children):
        l, name, attrs, s, r = children
        return [ep(l), name, attrs, w(s), ep(r)]

    def visit_fry_paired_element(self, node, children):
        return children

    def visit_fry_start_tag(self, node, children):
        l, name, attrs, s, r = children
        return [ep(l), name, attrs, w(s), ep(r)]

    def visit_fry_end_tag(self, node, children):
        l, name, s, r = children
        return [ep(l), name, w(s), ep(r)]

    def visit_fry_element_name(self, node, children):
        return en(node.text)

    def visit_fry_void_element_name(self, node, children):
        return en(node.text)

    def visit_space(self, node, children):
        return node.text

    def visit_maybe_space(self, node, children):
        return node.text

    def visit_fry_attributes(self, node, children):
        return children

    def visit_fry_spaced_attribute(self, node, children):
        s, attr = children
        return [w(s), attr]

    def visit_fry_attribute(self, node, children):
        return children[0]

    def visit_same_name_attribute(self, node, children):
        l, s1, identifier, s2, r = children
        return [sep(l), py(s1), py(identifier), py(s2), sep(r)]

    def visit_py_identifier(self, node, children):
        return node.text

    def visit_fry_embed_spread_attribute(self, node, children):
        l, s1, m, s2, script, r = children
        return [sep(l), py(s1), py(m), py(s2), script, sep(r)]

    def visit_fry_kv_attribute(self, node, children):
        name, s1, e, s2, value = children
        return [name, w(s1), o(e), w(s2), value]

    def visit_fry_novalue_attribute(self, node, children):
        return children[0]

    def visit_fry_attribute_name(self, node, children):
        return an(node.text)

    def visit_fry_attribute_value(self, node, children):
        return children[0]

    def visit_fry_children(self, node, children):
        return children

    def visit_fry_child(self, node, children):
        return children[0]

    #def visit_fry_js_embed(self, node, children):
    #    fry_embed, s, js_embed = children
    #    return [fry_embed, w(s), js_embed]

    def visit_joint_embed(self, node, children):
        fs, s, js = children
        return [fs, w(s), js]
        
    def visit_bracket_f_string(self, node, children):
        lb, body, rb = children
        return [sep(lb), fs(body), sep(rb)]

    def visit_bracket_f_string_body(self, node, children):
        return node.text

    def visit_single_f_string(self, node, children):
        lb, body, rb = children
        return [sep(lb), fs(body), sep(rb)]

    def visit_single_f_string_body(self, node, children):
        return node.text

    def visit_double_f_string(self, node, children):
        lb, body, rb = children
        return [sep(lb), fs(body), sep(rb)]

    def visit_double_f_string_body(self, node, children):
        return node.text

    def visit_fry_text(self, node, children):
        return t(node.text)

    def visit_no_embed_char(self, node, children):
        return t(node.text)

    def visit_maybe_web_script(self, node, children):
        return children

    def visit_web_script(self, node, children):
        s1, ls, attrs, s2, r1, script, r2 = children
        return [w(s1), ep('<'), en('script'), attrs, w(s2), ep(r1), script, ep('</'), en('script'), ep('>')]

    def visit_html_comment(self, node, children):
        return (Comment.HtmlComment, node.text)

    def visit_js_script(self, node, children):
        return js(node.text)

    def visit_js_embed(self, node, children):
        l, script, r = children
        return [cep(l), script, cep(r)]


class FStringLexer(PythonLexer):
    tokens = {
        'btdqf': [
            include('rfstringescape'),
            include('tdqf'),
        ],
        'btsqf': [
            include('rfstringescape'),
            include('tsqf'),
        ],
    }


class FryLexer(Lexer):
    name = 'Fry'
    aliases = ['fry']
    filenames = ['*.fry']
    mimetypes = ['text/fry']

    pylexer = PythonLexer()
    fslexer = FStringLexer()
    jslexer = JavascriptLexer()
    visitor = FryVisitor()

    def get_tokens_unprocessed(self, text):
        tree = grammar.parse(text)
        i = 0
        for t, v in self.visitor.visit(tree):
            if t in Token:
                yield i, t, v
            elif t == 'python':
                for i1, t1, v1 in self.pylexer.get_tokens_unprocessed(v):
                    yield i+i1, t1, v1
            elif t == 'fstring':
                root = 'btsqf' if '"""' in v else 'btdqf'
                for i1, t1, v1 in self.fslexer.get_tokens_unprocessed(v, stack=(root,)):
                    yield i+i1, t1, v1
            elif t == 'javascript':
                for i1, t1, v1 in self.jslexer.get_tokens_unprocessed(v):
                    yield i+i1, t1, v1
            else:
                raise ValueError(f"Invalid type '{t}': '{v}'.")
            i += len(v)


if __name__ == '__main__':
    import sys
    from pygments.formatters.terminal import TerminalFormatter
    from pygments import highlight
    lexer = FryLexer()
    fmter = TerminalFormatter()
    if len(sys.argv) > 1:
        with open(sys.argv[1], 'r', encoding='utf-8') as f:
            data = f.read()
        #print(highlight(data, lexer, fmter))
        for t, v in lexer.get_tokens(data):
            print(t, v)



from http import HTTPStatus
from django.http import Http404, HttpResponse, StreamingHttpResponse, JsonResponse
from django.conf import settings

from fryhcs.reload import event_stream, mime_type

from fryhcs import render

import logging
import uuid
import json

logger = logging.getLogger('frycss.views')


def check_hotreload(request):
    if not settings.DEBUG:
        raise Http404()
    if not request.accepts(mime_type):
        return HttpResponse(status=HTTPStatus.NOT_ACCEPTABLE)
    response = StreamingHttpResponse(
        event_stream(),
        content_type=mime_type,
    )
    response['content-encoding'] = ''
    return response

def component(request):
    if request.method == 'GET':
        data = request.GET
    elif request.method == 'POST':
        data = request.POST
    else:
        raise Http404()
    name = data.get('name')
    args = data.get('args')
    args = json.loads(args)
    element = render(name, **args)
    page = element.page
    return JsonResponse({
        'code': 0,
        'dom': str(element),
        'components': page.components,
    })



html_global_attributes = set([
    'accesskey',
    'autocapitalize',
    'autofocus',
    'contenteditable',
    'dir',
    'draggable',
    'enterkeyhint',
    'hidden',
    'inert',
    'inputmode',
    'is',
    'itemid',
    'itemprop',
    'itemref',
    'itemscope',
    'itemtype',
    'lang',
    'nonce',
    'popover',
    'spellcheck',
    'style',
    'tabindex',
    'title',
    'translate',
    'id',
    'class',
    'slot',
    'style',
    'role', # for aria
    ])

html_element_attributes = set([
    ('a',          ('href', 'target', 'download', 'ping', 'rel', 'hreflang', 'type', 'referrerpolicy',)),
    ('attr',       ()),
    ('address',    ()),
    ('area',       ('alt', 'coords', 'shape', 'href', 'target', 'download', 'ping', 'rel', 'referrerpolicy',)),
    ('article',    ()),
    ('aside',      ()),
    ('audio',      ('src', 'crossorigin', 'preload', 'autoplay', 'loop', 'muted', 'controls',)),
    ('b',          ()),
    ('base',       ('href', 'target',)),
    ('bdi',        ()),
    ('bdo',        ()),
    ('blockquote', ('cite',)),
    ('body',       ()),
    ('br',         ()),
    ('button',     ('disabled', 'form', 'formaction', 'formenctype', 'formmethod', 'formnovalidate', 'formtarget', 'name', 'popovertarget', 'popovertargetaction', 'type', 'value',)),
    ('canvas',     ('width', 'height',)),
    ('caption',    ()),
    ('cite',       ()),
    ('code',       ()),
    ('col',        ('span',)),
    ('colgroup',   ('span',)),
    ('data',       ('value',)),
    ('datalist',   ()),
    ('dd',         ()),
    ('del',        ('cite', 'datetime',)),
    ('details',    ('open',)),
    ('dfn',        ()),
    ('dialog',     ('open',)),
    ('div',        ()),
    ('dl',         ()),
    ('em',         ()),
    ('embed',      ('src', 'type', 'width', 'height',)),
    ('fieldset',   ('disabled', 'form', 'name',)),
    ('figcaption', ()),
    ('figure',     ()),
    ('footer',     ()),
    ('form',       ('accept-charset', 'action', 'autocomplete', 'enctype', 'method', 'name', 'novalidate', 'rel', 'target',)),
    ('h1',         ()),
    ('h2',         ()),
    ('h3',         ()),
    ('h4',         ()),
    ('h5',         ()),
    ('h6',         ()),
    ('head',       ()),
    ('header',     ()),
    ('hgroup',     ()),
    ('hr',         ()),
    ('html',       ('maifest',)),
    ('i',          ()),
    ('iframe',     ('src', 'srcdoc', 'name', 'sandbox', 'allow', 'allowfullscreen', 'width', 'height', 'referrerpolicy', 'loading',)),
    ('img',        ('alt', 'src', 'srcset', 'sizes', 'crossorigin', 'usemap', 'ismap', 'width', 'height', 'referrerpolicy', 'decoding', 'loading', 'fetchpriority',)),
    ('input',      ('accept', 'alt', 'autocomplete', 'checked', 'dirname', 'disabled', 'form', 'formaction', 'formenctype', 'formmethod', 'formnovalidate', 'formtarget', 'height', 'list', 'max', 'maxlength', 'min', 'minlength', 'multiple', 'name', 'pattern', 'placeholder', 'popovertarget', 'popovertargetaction', 'readonly', 'required', 'size', 'src', 'step', 'type', 'value', 'width',)),
    ('ins',        ('cite', 'datetime',)),
    ('kbd',        ()),
    ('label',      ('for',)),
    ('legend',     ()),
    ('li',         ('value',)),
    ('link',       ('href', 'crossorigin', 'rel', 'as', 'media', 'hreflang', 'type', 'sizes', 'imagesrcset', 'imagesizes', 'referrerpolicy', 'integrity', 'blocking', 'color', 'disabled', 'fetchpriority',)),
    ('main',       ()),
    ('map',        ('name',)),
    ('mark',       ()),
    ('math',       ()),
    ('menu',       ()),
    ('meta',       ('name', 'http-equiv', 'content', 'charset', 'media')),
    ('meter',      ('value', 'min', 'max', 'low', 'high', 'optimum',)),
    ('nav',        ()),
    ('noscript',   ()),
    ('object',     ('data', 'type', 'name', 'form', 'width', 'height',)),
    ('ol',         ('reversed', 'start', 'type')),
    ('optgroup',   ('disabled', 'label')),
    ('option',     ('disabled', 'label', 'selected', 'value')),
    ('output',     ('for', 'form', 'name')),
    ('p',          ()),
    ('picture',    ()),
    ('pre',        ()),
    ('progress',   ('value', 'max')),
    ('q',          ('cite',)),
    ('rp',         ()),
    ('rt',         ()),
    ('ruby',       ()),
    ('s',          ()),
    ('samp',       ()),
    ('script',     ('src', 'type', 'nomodule', 'async', 'defer', 'crossorigin', 'integrity', 'referrerpolicy', 'blocking', 'fetchpriority',)),
    ('search',     ()),
    ('section',    ()),
    ('select',     ('autocomplete', 'disabled', 'form', 'multiple', 'name', 'required', 'size')),
    ('slot',       ('name')),
    ('small',      ()),
    ('source',     ('type', 'media', 'src', 'srcset', 'sizes', 'width', 'height',)),
    ('span',       ()),
    ('strong',     ()),
    ('style',      ('media', 'blocking')),
    ('sub',        ()),
    ('summary',    ()),
    ('sup',        ()),
    ('svg',        ()),
    ('table',      ()),
    ('tbody',      ()),
    ('td',         ('colspan', 'rowspan', 'headers')),
    ('template',   ()),
    ('textarea',   ('autocomplete', 'cols', 'dirname', 'disabled', 'form', 'maxlength', 'minlength', 'name', 'placeholder', 'readonly', 'required', 'rows', 'wrap')),
    ('tfoot',      ()),
    ('th',         ('colspan', 'rowspan', 'headers', 'scope', 'abbr')),
    ('thead',      ()),
    ('time',       ('datetime')),
    ('title',      ()),
    ('tr',         ()),
    ('track',      ('default', 'kind', 'label', 'src', 'srclang',)),
    ('u',          ()),
    ('ul',         ()),
    ('var',        ()),
    ('video',      ('src', 'crossorigin', 'poster', 'preload', 'autoplay', 'playsnline', 'loop', 'muted', 'controls', 'width', 'height',)),
    ('wbr',        ()),
    ])

void_elements = set([
    'area',
    'base',
    'br',
    'col',
    'embed',
    'hr',
    'img',
    'input',
    'link',
    'meta',
    'source',
    'track',
    'wbr',
])

svg_elements = set([
    #'a',
    'animate',
    'animateMotion',
    'animateTransform',
    #'audio',
    #'canvas',
    'circle',
    'clipPath',
    'defs',
    'desc',
    'discard',
    'ellipse',
    'feBlend',
    'feColorMatrix',
    'feComponentTransfer',
    'feComposite',
    'feConvolveMatrix',
    'feDiffuseLighting',
    'feDisplacementMap',
    'feDistantLight',
    'feDropShadow',
    'feFlood',
    'feFuncA',
    'feFuncB',
    'feFuncG',
    'feFuncR',
    'feGaussianBlur',
    'feImage',
    'feMerge',
    'feMergeNode',
    'feMorphology',
    'feOffset',
    'fePointLight',
    'feSpecularLighting',
    'feSpotLight',
    'feTile',
    'feTurbulence',
    'filter',
    'foreignObject',
    'g',
    #'iframe',
    'image',
    'line',
    'linearGradient',
    'marker',
    'mask',
    'metadata',
    'mpath',
    'path',
    'pattern',
    'polygon',
    'polyline',
    'radialGradient',
    'rect',
    #'script',
    'set',
    'stop',
    #'style',
    'svg',
    'switch',
    'symbol',
    'text',
    'textPath',
    'title',
    'tspan',
    'unknown',
    'use',
])

html_elements = set(e for e, _ in html_element_attributes)

html_element_attribute_pairs = set((e, a) for e, al in html_element_attributes for a in al)

html_events = set([
    'DOMContentLoaded',
    'afterprint',
    'beforeprint',
    'beforematch',
    'beforetoggle',
    'beforeunload',
    'blur',
    'cancel',
    'change',
    'click',
    'close',
    'connect',
    'contextlost',
    'contextrestored',
    'currententrychange',
    'dispose',
    'error',
    'focus',
    'formdata',
    'hashchange',
    'input',
    'invalid',
    'languagechange',
    'load',
    'message',
    'messageerror',
    'navigate',
    'navigateerror',
    'navigatesuccess',
    'offline',
    'online',
    'open',
    'pagehide',
    'pageshow',
    'pointercancel',
    'popstate',
    'readystatechange',
    'rejectionhandled',
    'reset',
    'select',
    'storage',
    'submit',
    'toggle',
    'unhandledrejection',
    'unload',
    'visibilitychange',
    ])


def is_void_element(e):
    return e in void_elements

def is_valid_html_event(e):
    return e in html_events

def is_valid_html_element(e):
    return e in html_elements or '-' in e

def is_valid_html_attribute(e, a):
    return (a in html_global_attributes  or
            a[:5] in ('data-', 'aria-')  or
            '-' in e                     or
            e in svg_elements            or
            e in ('math', )              or
            (e, a) in html_element_attribute_pairs
           )

def normalize_html_attribute(e, a, v):
    if is_valid_html_attribute(e, a):
        return a
    return 'data-@' + a.replace('/', '@')



import uuid
import time

server_id = uuid.uuid4().hex

def update_serverid():
    global server_id
    server_id = uuid.uuid4().hex

def message():
    return f'data: {{"serverId": "{server_id}"}}\n\n'

def event_stream():
    # 立刻发送一个响应消息，触发浏览器EventSource的open事件。
    yield message()
    while True:
        time.sleep(0.5)
        # 定期发送消息，两个用途：
        # 1. 检测浏览器页面是否已关闭，关闭则结束这个链接
        # 2. 告诉浏览器自己还活着
        yield message()

mime_type = 'text/event-stream'



from collections import defaultdict
from pathlib import Path

class FileIter():
    def __init__(self, input_files=[]):
        self.path_globs = defaultdict(set)
        self.files = set()
        for file in input_files:
            if isinstance(file, str):
                self.add_file(file)
            elif isinstance(file, (tuple, list)):
                dir = file[0]
                for glob in file[1:]:
                    self.add_glob(dir, glob)

    def add_glob(self, path, glob):
        path = Path(path)
        try:
            path = path.absolute()
        except FileNotFoundError:
            return
        self.path_globs[path].add(glob)

    def add_file(self, file):
        path = Path(file)
        try:
            path = path.absolute()
        except FileNotFoundError:
            return
        self.files.add(path)

    def all_files(self):
        seen_files = set()
        for path, globs in self.path_globs.items():
            for glob in globs:
                for file in path.glob(glob):
                    if file in seen_files:
                        continue
                    seen_files.add(file)
                    yield file
        for file in self.files:
            if file in seen_files:
                continue
            seen_files.add(file)
            yield file



from parsimonious import BadGrammar
from pathlib import Path
import sys
from fryhcs.fry.grammar import grammar
from fryhcs.fry.generator import BaseGenerator
from fryhcs.fileiter import FileIter
from fryhcs.element import ref_attr_name, refall_attr_name
import re
import os
import subprocess
import shutil


# generate js content for fry component
# $component：代表组件对象
# embeds： js嵌入值列表
def compose_js(args, script, embeds):
    output = []
    if args:
        args = f'let {{ {", ".join(args)} }} = $component.fryargs;'
    else:
        args = ''

    return f"""\
export {{ hydrate }} from "fryhcs";
export const prepare = async function ($component) {{
    let $fryobject = undefined;
    {args}
    {script}
    Object.assign($component, $fryobject);
    $component.fryembeds = [{', '.join(embeds)}];
}};
"""


class JSGenerator(BaseGenerator):
    component_pattern = f"**/[A-Z]*-{'[0-9a-fA-Z]'*40}.js"

    def __init__(self, input_files, output_dir):
        super().__init__()
        self.fileiter = FileIter(input_files)
        self.output_dir = Path(output_dir).absolute()
        self.tmp_dir = self.output_dir / '.tmp'

    def generate(self, input_files=[], clean=False):
        if not input_files:
            input_files = self.fileiter.all_files()
        self.output_dir.mkdir(parents=True, exist_ok=True)
        self.tmp_dir.mkdir(parents=True, exist_ok=True)
        if clean:
            for f in self.output_dir.glob(self.component_pattern):
                f.unlink(missing_ok=True)
            for f in self.tmp_dir.glob('*.js'):
                f.unlink(missing_ok=True)
        self.dependencies = set()
        count = 0
        for file in input_files:
            self.set_curr_file(file)
            self.js_dir = self.tmp_dir / self.relative_dir
            self.js_dir.mkdir(parents=True, exist_ok=True)
            # 设置newline=''确保在windows下换行符为\r\n，文件内容不会被open改变
            # 参考[universal newlines mode](https://docs.python.org/3/library/functions.html#open-newline-parameter)
            with self.curr_file.open('r', encoding='utf-8', newline='') as f:
                count += self.generate_one(f.read())
        self.bundle()
        return count
                
    def bundle(self):
        if self.dependencies:
            for file, path in self.dependencies:
                p = self.tmp_dir / path
                p.mkdir(parents=True, exist_ok=True)
                shutil.copy(file, p)
        src = list(self.tmp_dir.glob(self.component_pattern))
        if not src:
            return
        this = Path(__file__).absolute().parent
        bun = this / 'bun' 
        env = os.environ.copy()
        if True:
            # 2024.2.23: bun不成熟，使用esbuild打包
            # esbuild支持通过环境变量NODE_PATH设置import查找路径
            env['NODE_PATH'] = str(this / '..' / 'static' / 'js')
            # Windows上需要指定npx全路径，否则会出现FileNotFoundError
            npx = shutil.which('npx')
            if not npx:
                print(f"Can't find npx, please install nodejs first.")
                return
            args = [npx, 'esbuild', '--format=esm', '--bundle', f'--outbase={self.tmp_dir}']
        elif bun.is_file():
            # bun的问题：对于动态import的js，只修改地址，没有打包
            # 暂时不用bun
            args = [str(bun), 'build', '--external', 'fryhcs']
        args += ['--splitting', f'--outdir={self.output_dir}']
        args += [str(js) for js in src]
        subprocess.run(args, env=env)

    def check_js_module(self, jsmodule):
        if jsmodule[0] in "'\"":
            jsmodule = jsmodule[1:-1]
        if jsmodule.startswith('./') or jsmodule.startswith('../'):
            jsfile = self.curr_dir / jsmodule
            jsfile = jsfile.resolve(strict=True) # 如果js文件不存在，抛出异常
            jsdir = jsfile.parent.relative_to(self.curr_root)
            self.dependencies.add((str(jsfile), str(jsdir)))

    def generate_one(self, source):
        tree = grammar.parse(source)
        self.web_components = []
        self.script = ''
        self.args = []
        self.embeds = []
        self.refs = set()
        self.refalls = set()
        self.visit(tree)
        for c in self.web_components:
            name = c['name']
            args = c['args']
            script = c['script']
            embeds = c['embeds']
            jspath = self.js_dir / f'{name}.js'
            with jspath.open('w', encoding='utf-8') as f:
                f.write(compose_js(args, script, embeds))
        return len(self.web_components)

    def generic_visit(self, node, children):
        return children or node

    def visit_single_quote(self, node, children):
        return node.text

    def visit_double_quote(self, node, children):
        return node.text

    def visit_py_simple_quote(self, node, children):
        return children[0]

    def visit_js_simple_quote(self, node, children):
        return children[0]

    def visit_fry_component(self, node, children):
        cname, _fryscript, _template, _script = children
        if self.script or self.embeds or self.refs or self.refalls:
            uuid = self.get_uuid(cname, node)
            self.web_components.append({
                'name': uuid,
                'args': [*self.refs, *self.refalls, *self.args],
                'script': self.script,
                'embeds': self.embeds})
        self.script = ''
        self.args = []
        self.embeds = []
        self.refs = set()
        self.refalls = set()

    def visit_fry_component_header(self, node, children):
        _def, _, cname, _ = children
        return cname

    def visit_fry_component_name(self, node, children):
        return node.text

    def visit_fry_attributes(self, node, children):
        return [ch for ch in children if ch]

    def visit_fry_spaced_attribute(self, node, children):
        _, attr = children
        return attr

    def visit_fry_attribute(self, node, children):
        return children[0]

    def visit_same_name_attribute(self, node, children):
        _l, _, identifier, _, _r = children
        return identifier

    def visit_py_identifier(self, node, children):
        return node.text

    def visit_fry_embed_spread_attribute(self, node, children):
        return None

    def visit_fry_kv_attribute(self, node, children):
        name, _, _, _, value = children
        name = name.strip()
        if name == ref_attr_name:
            _type, script = value
            value = script.strip()
            if value in self.refs or value in self.refalls:
                raise BadGrammar(f"Duplicated ref name '{value}', please use 'refall'")
            self.refs.add(value)
            return None
        elif name == refall_attr_name:
            _type, script = value
            value = script.strip()
            if value in self.refs:
                raise BadGrammar(f"Ref name '{value}' exists, please use another name for 'refall'")
            self.refalls.add(value)
            return None
        elif isinstance(value, tuple) and value[0] == 'js_embed':
            script = value[1]
            self.embeds.append(script)
        return name

    def visit_fry_novalue_attribute(self, node, children):
        return children[0]

    def visit_fry_attribute_name(self, node, children):
        return node.text

    def visit_fry_attribute_value(self, node, children):
        return children[0]

    def visit_joint_embed(self, node, children):
        _f_string, _, jsembed = children
        _name, script = jsembed
        self.embeds.append(script)
        return None

    def visit_web_script(self, node, children):
        _, _begin, attributes, _, _greaterthan, script, _end = children
        self.args = [k for k in attributes if k]
        self.script = script

    def visit_js_script(self, node, children):
        return ''.join(str(ch) for ch in children)

    def visit_js_embed(self, node, children):
        _, script, _ = children
        return ('js_embed', script)

    def visit_js_parenthesis(self, node, children):
        _, script, _ = children
        return '(' + script + ')'

    def visit_js_brace(self, node, children):
        _, script, _ = children
        return '{' + script + '}'

    def visit_js_script_item(self, node, children):
        return children[0]

    def visit_js_single_line_comment(self, node, children):
        return node.text

    def visit_js_multi_line_comment(self, node, children):
        return node.text

    def visit_js_regexp(self, node, children):
        return node.text

    def visit_js_template_simple(self, node, children):
        return node.text

    def visit_js_template_normal(self, node, children):
        return node.text

    def visit_js_static_import(self, node, children):
        return children[0]

    def visit_js_simple_static_import(self, node, children):
        _, _, module_name = children
        self.check_js_module(module_name)
        return f'await import({module_name})'

    def visit_js_normal_static_import(self, node, children):
        _import, _, identifiers, _, _from, _, module_name = children
        self.check_js_module(module_name)
        value = ''
        namespace = identifiers.pop('*', '')
        if namespace:
            value = f'const {namespace} = await import({module_name})'
            if identifiers:
                value += ', '
        names = []
        for k,v in identifiers.items():
            if v:
                names.append(f'{k}: {v}')
            else:
                names.append(k)
        if names:
            names = ", ".join(names)
            if namespace:
                value += f'{{{names}}} = {namespace}'
            else:
                value += f'const {{{names}}} = await import({module_name})'
        return value

    def visit_js_import_identifiers(self, node, children):
        identifier, others = children
        identifiers = identifier
        identifiers.update(others)
        return identifiers
        
    def visit_js_other_import_identifiers(self, node, children):
        identifiers = {}
        for ch in children:
            identifiers.update(ch)
        return identifiers

    def visit_js_other_import_identifier(self, node, children):
        _, _comma, _, identifier = children
        return identifier

    def visit_js_import_identifier(self, node, children):
        if isinstance(children[0], str):
            return {'default': children[0]}
        else:
            return children[0]

    def visit_js_identifier(self, node, children):
        return node.text

    def visit_js_namespace_import_identifier(self, node, children):
        _star, _, _as, _, identifier = children
        return {'*': identifier}

    def visit_js_named_import_identifiers(self, node, children):
        _lb, _, identifier, others, _, _rb = children
        identifiers = identifier
        identifiers.update(others)
        return identifiers

    def visit_js_other_named_import_identifiers(self, node, children):
        identifiers = {}
        for ch in children:
            identifiers.update(ch)
        return identifiers

    def visit_js_other_named_import_identifier(self, node, children):
        _, _comma, _, identifier = children
        return identifier

    def visit_js_named_import_identifier(self, node, children):
        value = children[0]
        if isinstance(value, str):
            return {value: ''}
        else:
            return value

    def visit_js_identifier_with_alias(self, node, children):
        identifier, _, _as, _, alias = children
        return {identifier: alias}

    def visit_js_default_export(self, node, children):
        return '$fryobject ='

    def visit_js_normal_code(self, node, children):
        return node.text

    def visit_no_script_less_than_char(self, node, children):
        return node.text

    def visit_no_comment_slash_char(self, node, children):
        return node.text

    def visit_no_import_i_char(self, node, children):
        return node.text

    def visit_no_export_e_char(self, node, children):
        return node.text



from django.apps import AppConfig


class FryhcsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'fryhcs'

    def ready(self):
        # 注册信号处理函数
        from . import signals

        # 让python可以import .fry文件
        from .fry.fryloader import install_path_hook
        install_path_hook()



from parsimonious import Grammar
from pathlib import Path

def load_grammar():
    grammar_file = Path(__file__).parent / 'fry.ppeg'
    with grammar_file.open('r', encoding='utf-8') as gf:
        grammar_text = gf.read()
    return Grammar(grammar_text)

grammar = load_grammar()



from pathlib import Path
import os

try:
    from django.conf import settings as django_settings
except ImportError:
    django_settings = False

try:
    from flask import current_app as flask_app
except ImportError:
    flask_app = False

class FryConfig():
    @property
    def django_ok(self):
        try:
            return django_settings and django_settings.configured
        except:
            pass
        return False

    @property
    def flask_ok(self):
        if flask_app:
            try:
                _current_obj = flask_app._get_current_object()
                return True
            except RuntimeError:
                pass
        return False

    def item(self, name, default):
        if self.django_ok:
            return getattr(django_settings, name, default)
        if self.flask_ok:
            return flask_app.config.get(name, default)
        if name in os.environ:
            value = os.environ[name]
            if isinstance(default, (list, tuple)):
                value = value.split(':')
            return value
        return default

    @property
    def js_url(self):
        return self.item('FRYHCS_JS_URL', 'js/components/')

    @property
    def css_url(self):
        return self.item('FRYHCS_CSS_URL', 'css/styles.css')

    @property
    def check_reload_url(self):
        if not self.debug:
            return ''
        if self.django_ok:
            from django.urls import reverse
            return reverse('fryhcs:check_hotreload')
        if self.flask_ok:
            return flask_app.url_for('fryhcs_check_hotreload')
        return ''

    @property
    def debug(self):
        return self.item('DEBUG', True)

    @property
    def static_root(self):
        if self.django_ok:
            root_dir = getattr(django_settings, 'FRYHCS_STATIC_ROOT', '')
            if not root_dir:
                staticfiles_dirs = getattr(django_settings, 'STATICFILES_DIRS', [])
                if len(staticfiles_dirs) == 0:
                    raise Exception("'FRYHCS_STATIC_ROOT' is not specified, 'STATICFILES_DIRS' should have at least one item.")
                root_dir = staticfiles_dirs[0]
            return Path(root_dir)
        if self.flask_ok:
            return Path(flask_app.static_folder)
        return Path('.')

    @property
    def config_root(self):
        root_dir = None
        if self.django_ok:
            root_dir = getattr(django_settings, 'BASE_DIR', '')
        if self.flask_ok:
            root_dir = flask_app.root_path
        if root_dir:
            return Path(root_dir).resolve()
        else:
            return Path('.').resolve()

    @property
    def semantic_theme(self):
        return self.item('FRYHCS_SEMANTIC_THEME', None)

    @property
    def plugins(self):
        return self.item('FRYHCS_PLUGINS', [])

    @property
    def static_url(self, default='/static'):
        if self.django_ok:
            return getattr(django_settings, 'STATIC_URL', default)
        if self.flask_ok:
            return flask_app.static_url_path
        return default

    @property
    def js_root(self):
        return self.static_root / self.js_url

    @property
    def css_file(self):
        return self.static_root / self.css_url

fryconfig = FryConfig()



from fryhcs.config import fryconfig
from fryhcs.css.style import CSS
from importlib import import_module
import re

class PluginError(Exception):
    pass

base_csses = []
static_utilities = {}
dynamic_utilities = {}
semantic_colors = {}

def load_plugins():
    for pid, plugin in enumerate(fryconfig.plugins):
        module = import_module(plugin)
        if module:
            load_plugin(pid, module)

def load_plugin(pid, plugin):
    base_css = getattr(plugin, 'base_css', lambda:{})()
    utilities = getattr(plugin, 'utilities', lambda:{})()
    #types = getattr(plugin, 'types', lambda:{})()
    colors = getattr(plugin, 'colors', lambda:{})()

    if base_css:
        base_csses.append(base_css)

    if colors:
        semantic_colors.update(colors)

    level = 0
    for name, content in prepare_utilities(utilities):
        #load_utility(name, content, types, pid, level)
        load_utility(name, content, pid, level)
        level += 1


def prepare_utilities(utilities):
    newlist = []
    newdict = {}
    def add(n, v):
        n = n.strip()
        if ':' in n:
            raise RuntimeError(f"Custom utility should not has modifiers: '{n}'")
        if not n in newdict:
            newdict[n] = v
            newlist.append((n,v))
        else:
            newdict[n] += v
    for name, *value in utilities:
        if ',' in name:
            for n in name.split(','):
                add(n, value[:])
        else:
            add(name, value)
    return newlist


# 2023.12.16 新定义的utility名中暂不支持变量，有了colors后，变量意义已经不大了。
## utility1: "btn-<color:state-color>-focus"
## return1:  "btn-(info|success|warning|error)-focus", [("color", lambda v: f"var(--{v})")]
## utility2: "btn"
## return2:  "btn", []
#def parse_name(utility, types):
#    r = '<([0-9a-zA-Z_-]+)(?::([0-9a-zA-Z_-]+))?>'
#    vs = []
#    lb = 0
#    names = [] 
#    for m in re.finditer(r, utility):
#        begin, end = m.span()
#        name, t = m.groups()
#        if t is None:
#            t = 'DEFAULT'
#        tt = types[t]
#        names.append(utility[lb:begin])
#        names.append('(')
#        names.append(tt['re'])
#        names.append(')')
#        vs.append((name, tt['value']))
#        lb = end
#    names.append(utility[lb:])
#    return (''.join(names), vs)


def prepare_content(content):
    newlist = []
    for name, *value in content:
        if ',' in name:
            for n in name.split(','):
                newlist.append((n.strip(), value))
        else:
            newlist.append((name.strip(), value))
    return newlist


def load_subutilities(prefix, content):
    dummy = None
    subcsses = []
    for key, *value in content:
        if key == '@apply':
            dummy = None
            if not all(isinstance(v, str) for v in value):
                raise PluginError(f"@apply can only have string values, not '{value}'.")
            value = ' '.join(value)
            subutils = value.split()
            subcsses += [CSS(key=prefix, value=v) for v in subutils]
        elif '&' in key or ',' in key:
            raise PluginError(f"Subutilities can't have this kind of key: '{key}'")
        elif len(value) == 1 and isinstance(value[0], str):
            if not dummy:
                dummy = CSS(key=prefix, value='dummy')
                subcsses.append(dummy)
            dummy.add_style(key, value[0])
        else:
            raise PluginError(f"Invalid '{name}': '{value}'")
    return subcsses

#def load_utility(name, content, types, pid, level):
def load_utility(name, content, pid, level):
    #regexp, variables = parse_name(name, types)
    content = prepare_content(content)
    dummy = None
    defaultcss = CSS()
    #defaultcss.selector = regexp
    #defaultcss._variables = variables
    defaultcss.selector = name
    defaultcss.plugin_order = pid
    defaultcss.level_order = level
    subcsses = []
    for key, value in content:
        if key == '@apply':
            dummy = None
            if not all(isinstance(v, str) for v in value):
                raise PluginError(f"@apply can only have string values, not '{value}'.")
            value = ' '.join(value)
            subutils = value.split()
            subcsses += [CSS(value=v) for v in subutils]
        elif key.endswith(':&'):
            dummy = None
            if '&' in key[:-2]:
                raise PluginError(f"Invalid format for '{key}'.")
            subcsses += load_subutilities(key[:-2], value)
        elif len(value) == 1 and isinstance(value[0], str):
            if not dummy:
                dummy = CSS()
                subcsses.append(dummy)
            dummy.add_style(key, value[0])
        else:
            raise PluginError(f"Invalid '{name}': '{value}'")
    csses = defaultcss.addons
    lastkey = None
    for css in subcsses:
        key = (css.selector_template, *css.wrappers)
        if key == lastkey:
            csses[-1].styles += css.styles
        else:
            csses.append(css)
            lastkey = key
    static_utilities[defaultcss.selector] = defaultcss


def plugin_utility(utility_args):
    utility = '-'.join(utility_args)
    if utility in static_utilities:
        clone = static_utilities[utility].clone()
        return clone
    # 2023.12.16 暂不支持动态utility，有了colors后动态utility意义已经不大了
    #for regexp in dynamic_utilities.keys():
    #    # 动态utility可能有负号，负号被放到utility之前，处理负号的情况
    #    negative = ''
    #    if utility[0] == '-':
    #        negative = '-'
    #        utility = utility[1:]
    #    match = re.fullmatch(regexp, utility)
    #    if match:
    #        du = dynamic_utilities[regexp]
    #        values = match.groups()
    #        args = {'NEGATIVE': negative}
    #        for val, var in zip(values, du._variables):
    #            args[var[0]] = var[1](val)
    #        clone = du.clone()
    #        clone.update_values(args)
    #        return clone
    return None 

def plugin_color(arg):
    return semantic_colors.get(arg, None)

def plugin_basecss():
    output = []
    def lines(base, indent):
        for key, value in base.items():
            if isinstance(value, (str, int)):
                output.append(f'{indent}{key}: {value};')
            elif isinstance(value, (list, tuple)):
                output.append(f'{indent}{key}: {" ".join(str(x) for x in value)};')
            elif isinstance(value, dict):
                output.append(f'{indent}{key} {{')
                lines(value, indent+'  ')
                output.append(f'{indent}}}')

    for basecss in base_csses:
        lines(basecss, '')
    return '\n'.join(output)+'\n'

load_plugins()



# based on django_browser_reload

from pathlib import Path

from django.dispatch import receiver
from django.utils.autoreload import autoreload_started, file_changed, is_django_path
from django.contrib.staticfiles.finders import get_finders

from fryhcs.config import fryconfig

import threading
import logging

RELOAD_DEBOUNCE_TIME = 0.05

browser_reload_event = threading.Event()
browser_reload_timer = None

css_generator = None
js_generator = None

incremental_generation_count = 0

logger = logging.getLogger('fryhcs.signals')

def trigger_browser_reload():
    browser_reload_event.set()

def browser_reload_soon():
    global browser_reload_timer
    if browser_reload_timer is not None:
        browser_reload_timer.cancel()
    browser_reload_timer = threading.Timer(RELOAD_DEBOUNCE_TIME, trigger_browser_reload)
    browser_reload_timer.start()


def staticfiles_storages():
    for finder in get_finders():
        if hasattr(finder, 'storages') and isinstance(finder.storages, dict):
            yield from finder.storages.values()


@receiver(autoreload_started, dispatch_uid='fryhcs_start_generator')
def start_generator(sender, **kwargs):
    # django框架只监控了django模板的变更，为了浏览器自动更新，
    # 添加对静态资源的监控
    for storage in staticfiles_storages():
        if hasattr(storage, 'location'):
            sender.watch_dir(Path(storage.location), '**/*')

    from fryhcs.utils import create_css_generator, create_js_generator
    global js_generator
    global css_generator
    js_generator = create_js_generator()
    js_generator.generate()
    css_generator = create_css_generator()
    css_generator.generate()


@receiver(file_changed, dispatch_uid='fryhcs_file_changed')
def template_changed(sender, file_path, **kwargs):
    global incremental_generation_count
    if is_django_path(file_path):
        # 我处理不了，不行就重启吧（服务端reload）
        return

    if file_path.is_dir():
        # 目录变更，不触发服务端reload
        return True

    if file_path.suffix == '.swp':
        # vim编辑临时文件，不触发服务端reload
        return True

    # 当有fry/pyx文件发生变化时，更新css/js文件
    if file_path.suffix in ('.fry', '.pyx'):
        logger.info("generate css for %s...", file_path)
        incremental_generation_count += 1
        if incremental_generation_count >= 10:
            # 为减少垃圾样式信息，每10次增量生成后，做一次全量生成
            incremental_generation_count = 0
            css_generator.generate()
        else:
            css_generator.generate(file_path)
        logger.info("generate js for %s...", file_path)
        js_generator.generate([file_path])

    # 当有html、js或css文件发生变更时，不需要服务端reoad，但浏览器需要reload
    if file_path.suffix in ('.html', '.js', '.css'):
        from fryhcs.reload import update_serverid
        update_serverid()
        return True
    # 其他情况，可能要服务端reload
    else:
        return



print('happybirthday plugin')

base_css = {
    '@keyframes explosion': {
        '0%': {
            'opacity': 0,
        },
        '70%': {
            'opacity': 1,
        },
        '100%': {
            'transform': 'translate(50vw, 100vh)',
        },
    },
    '@keyframes flicker': {
        '0%': {
            'transform': 'skewX(5deg)',
        },
        '25%': {
            'transform': 'skewX(-5deg)',
        },
        '50%': {
            'transform': 'skewX(10deg)',
        },
        '75%': {
            'transform': 'skewX(-10deg)',
        },
        '100%': {
            'transform': 'skewX(5deg)',
        },
    },
}



import re

from .modifiers import is_modifier, add_modifier
from .utilities import Utility


def quote_selector(value):
    invalid_chars = r"([~!@$%^&*()+=,./';:\"?><[\]\\{}|`#])"
    fun = lambda m: '\\'+m.group(1)
    return re.sub(invalid_chars, fun, value)


class CSS():
    default_selector_template = '{selector}'
    def __init__(self, key='', value='', toclass=True, generate=True):
        self.key = key
        self.value = value
        self.toclass = toclass
        self.wrappers = []
        self.modifiers = []
        self.utility_args = []
        self.selector = ''
        self.selector_template = self.default_selector_template
        self.styles = []
        self.addons = []
        self.plugin_order = 1000 # 默认最大，优先级最高
        self.level_order = 0
        self.screen_order = []
        self.valid = True
        self.parse()
        if generate:
            self.generate()

    def clone(self):
        css = CSS(generate=False)
        css.key = self.key
        css.value = self.value
        css.toclass = self.toclass
        css.modifiers = self.modifiers[:]
        css.utility_args = self.utility_args[:]
        css.selector = self.selector
        css.selector_template = self.selector_template
        css.styles = self.styles[:]
        css.addons = [addon.clone() for addon in self.addons]
        css.plugin_order = self.plugin_order
        css.level_order = self.level_order
        css.screen_order = self.screen_order
        css.valid = self.valid
        return css

    def update_values(self, args):
        self.styles = [(k, v.format_map(args)) for k,v in self.styles]
        for addon in self.addons:
            addon.update_values(args)

    @property
    def order(self):
        return (self.plugin_order, self.level_order, *self.screen_order)

    @classmethod
    def union(cls, csses):
        # TODO
        pass

    def to_class(self):
        modifiers = ':'.join(self.modifiers)
        utility = '-'.join(self.utility_args)
        if modifiers:
            if utility:
                return modifiers + ':' + utility
            else:
                return modifiers
        else:
            return utility

    def clean_args(self, args):
        # border="~ cyan-100"将被匹配为border和border-cyan-100
        # 而非border-~和border-cyan-100
        if len(args) > 1 and args[-1] == '~':
            return args[:-1]
        return args

    def parse(self):
        """
        for class, key = '';
        for no-value attribute, value = ''
        from:
            self.key
            self.value
        generate:
            self.selector: css selector based on the value of key and value
            self.modifiers: all modifiers
            self.utility_args: utility and its args
        如果utility中的大小为负值，则utility_args[0]以'-'开头
        """
        key = self.key
        value = self.value

        if not key and not value:
            self.modifiers = []
            self.utility_args = []
            self.selector = ''
            return

        if not key:
            # class的css匹配方式: .classname
            selector = '.' + quote_selector(value)
        elif not value:
            # 只有属性名没有值的匹配方式：[key=""]，不能用[key]，
            # [key]匹配的是存在属性key，此时key的值可以是任意值；
            # 也不能用[key ~= ""]，这个无法匹配只有属性名的情况。
            selector = '[' + quote_selector(key) + ' = ""]'
        else:
            # 既有属性又有值的情况，与类的情况类似，值中使用空格分开
            # 的每一个"子值"，都是用[key ~= subvalue]进行匹配
            selector = '[' + quote_selector(key) + ' ~= "' + value + '"]'

        keys = key.split(':') if key else []
        values = value.split(':') if value else []
        negative = False
        if keys and not is_modifier(keys[-1]):
            modifiers = keys[:-1]
            utility = keys[-1]
            if utility and utility[0] == '-':
                utility = utility[1:]
                negative = not negative
            if utility:
                utility_args = utility.split('-')
            else:
                utility_args = []
        else:
            modifiers = keys[:]
            utility_args = []
        if values:
            modifiers += values[:-1]
            utility = values[-1]
            if utility and utility[0] == '-':
                utility = utility[1:]
                negative = not negative
            if utility:
                utility_args += utility.split('-')
        if negative and utility_args:
            utility_args[0] = '-' + utility_args[0]

        self.modifiers = modifiers
        self.utility_args = self.clean_args(utility_args)
        self.selector = selector
        if self.toclass:
            self.selector = '.' + quote_selector(self.to_class())

    def generate(self):
        """
        from:
            self.modifiers
            self.utility_args
        generate:
            self.wrappers
            self.selector_template
            self.styles
            self.addons
        """
        for modifier in self.modifiers:
            if not add_modifier(self, modifier):
                self.valid = False
                return
        self.utility = Utility(self)
        self.valid = self.utility()
    
    def add_style(self, key, value):
        style = (key, value)
        self.styles.append(style)

    def get_style(self, key):
        for k, v in self.styles:
            if k == key:
                return v
        return None

    def new_addon(self, key='', value='', generate=False):
        addon = CSS(key, value, generate)
        self.addons.append(addon)
        return addon

    def add_addon(self, addon):
        self.addons.append(addon)

    def quote(self, value):
        return quote_selector(value)

    def has_styles(self):
        if not self.valid:
            return False
        if len(self.styles) > 0:
            return True
        return any(addon.has_styles() for addon in self.addons)

    def lines(self, new_selector=''):
        if not self.valid:
            return []
        twospace = '  '
        indent = ''
        lines = []
        if not self.has_styles():
            return lines
        for wrapper in self.wrappers:
            lines.append(indent + wrapper + ' {')
            indent += twospace
        selector = new_selector or self.selector
        selector = self.selector_template.format(selector=selector)
        if len(self.styles) > 0:
            lines.append(indent + selector + ' {')
            indent += twospace
            for k, v in dict(self.styles).items():
                style = f'{k}: {v};'
                lines.append(indent + style)
            indent = indent[:-2]
            lines.append(indent + '}')
        for addon in self.addons:
            lines.append('')
            for line in addon.lines(selector):
                lines.append(indent + line)
        while len(indent) > 0:
            indent = indent[:-2]
            lines.append(indent + '}')
        return lines

    def text(self):
        return '\n'.join(self.lines()) + '\n\n'



from pathlib import Path

from fryhcs.config import fryconfig

import os
import inspect
import sys

if fryconfig.django_ok:
    from django.template.autoreload import get_template_directories as django_template_directories
    def jinja_template_directories():
        try:
            from django.template.backends.jinja2 import Jinja2
            from django.template import engines
        except ImportError:
            return set()
        cwd = Path.cwd()
        items = set()
        for backend in engines.all():
            if not isinstance(backend, Jinja2):
                continue

            loader = backend.env.loader  # type: ignore [attr-defined]
            if hasattr(loader, searchpath) and isinstance(loader.searchpath, (list, tuple)):
                items.update([cwd / Path(fspath) for fspath in loader.searchpath])
        return items

    def template_directories():
        dirs = django_template_directories()
        dirs.update(jinja_template_directories())
        return dirs

def fry_files():
    paths = set()
    if fryconfig.django_ok:
        from django.apps import apps
        paths.update(Path(ac.path).resolve() for ac in apps.get_app_configs())
        input_files = [(ac.path, '**/*.fry') for ac in apps.get_app_configs()]
    elif fryconfig.flask_ok:
        from flask import current_app
        try:
            paths.add(Path(current_app.root_path).resolve())
            input_files = [(current_app.root_path, '**/*.fry')]
        except RuntimeError:
            pass
    syspath = [Path(p).resolve() for p in sys.path]
    paths.update(p.resolve() for p in syspath if p.is_dir())
    input_files = [(str(p), '**/*.fry') for p in paths]
    if not input_files:
        raise RuntimeError('django or flask is not configured')

    return input_files

def create_css_generator():
    # input_files = [(dir, '**/*.html') for dir in template_directories()]
    from fryhcs.css.generator import CSSGenerator
    return CSSGenerator(fry_files(), fryconfig.css_file)

def create_js_generator():
    from fryhcs.js.generator import JSGenerator
    return JSGenerator(fry_files(), fryconfig.js_root)


def static_url(path):
    if fryconfig.django_ok:
        from django.contrib.staticfiles.storage import staticfiles_storage
        return staticfiles_storage.url(path)
    elif fryconfig.flask_ok:
        from flask import current_app
        try:
            return current_app.static_url_path.rstrip('/') + '/' + path.lstrip('/')
        except RuntimeError:
            pass
    return path

def component_name(fn):
    if inspect.isfunction(fn) or inspect.isclass(fn):
        if fryconfig.django_ok:
            from django.apps import apps
            ac = apps.get_containing_app_config(fn.__module__)
            if ac:
                return ac.label + ':' + fn.__name__
        elif fryconfig.flask_ok:
            from flask import current_app
            try:
                name = current_app.name
                if name:
                    return name + ':' + fn.__name__
            except RuntimeError:
                pass
        return fn.__name__
    return str(fn)



"""
    fryhcs.__main__
    ~~~~~~~~~~~~~~~

    Main entry point for ``python -m fryhcs``

    :copyright: Copyright 2023 by zenkj<juzejian@gmail.com>
    :license: BSD, see LICENSE for details.
"""

import sys
from fryhcs.cmdline import main

try:
    sys.exit(main())
except KeyboardInterrupt:
    sys.exit(1)
